#!/usr/bin/env python
"""
Devapp Project Creation

By providing the --init_at <project dir> switch we will (re-)initialize a project directory there, incl:

- Installing available resources, like databases and tools within a given directory (conda_prefix)
- Creating resource start wrappers in <project_dir>/bin
- Creating a 

Privilege escalation is not required for any of these steps, your system remains unchanged, except <project_dir> and <conda_prefix>

"""
# Could be done far smaller.
from re import match
from devapp import gevent_patched
import hashlib
import json
import os
import shutil
from copy import deepcopy
from functools import partial

import requests
from devapp.app import FLG, app, run_app, do, system
from devapp import tools
from devapp.tools import (
    exists,
    to_list,
    repl_dollar_var_with_env_val,
    project,
    read_file,
    write_file,
)
import devapp.resource_tools as api


class Flags(api.CommonFlags):
    autoshort = ''

    class force:
        n = 'Assume y on all questions. Required when started w/o a tty'
        d = False

    #     class skip_inst_when_found:
    #         n = 'Do not install resources which are present on the host already'
    #         d = False
    class force_reinstall:
        n = 'Do not only install resources detected uninstalled but reinstall all'
        d = False

    class init_at:
        n = 'Set up project in given directory. env vars / relative dirs supported.'
        d = ''

    class init_create_all_units:
        n = 'If set we create unit files for ALL service type resources'
        d = False

    class init_create_unit_files:
        'Valid: Entries in rsc.provides, rsc.cmd or rsc.exe (i.e. the filename of the wrapper in bin dir). Not: rsc.name.'
        n = 'List service unit files you want to have created for systemctl --user.'
        d = []

    class init_resource_match:
        n = 'Install only matching resources. Examples: -irm "redis, hub" or -irm \'!mysql, !redis\' (! negates).'
        d = []

    class list_resources_files:
        n = 'Show available definition files.'
        d = False

    class add_post_process_cmd:
        n = 'Add this to all commands which have systemd service units. Intended for output redirection. Not applied when stdout is a tty.\n'
        n += '''Example: -appc='2>&1 | rotatelogs -e -n1 "$logfile" 1M' ($logfile defined in wrapper -> use single quotes).\n'''
        n += 'Tip: Use rotatelogs only on powers of 10 - spotted problems with 200M. Use 100M or 1G in that case.'
        d = ''

    class edit_matching_resource_file:
        n = 'Open resource files in $EDITOR, matching given string in their content'
        d = ''

    class delete_all_matching_service_unit_files:
        n = 'This removes all matching unit files calling devapp service wrappers. Say "service" to match all'
        d = ''


S = api.S


def create_project_dirs():
    ds = ['bin', 'data', 'log', 'work', 'conf', 'tmp', 'build']
    D = project.root()
    if not FLG.fs_dir:
        ds.append('fs')
    for d in ds:
        d = D + '/' + d
        if not exists(d):
            app.info('creating', dir=d)
            os.makedirs(d)
        else:
            app.debug('exists already', dir=d)


def git_init():
    D = project.root()
    if not exists(D + '/.git'):
        do(system, 'git init')
    fn = D + '/.gitignore'
    if not exists(fn):
        s = '\n'.join(['data/', 'log/', '*.py[cod]'])
        write_file(fn, s)


import sys

rscs_dicts = lambda rscs: [api.to_dict(r) for r in rscs]
rscs_details = lambda rscs: app.info('Resource details', json=rscs_dicts(rscs))


def confirm(d, rscs):
    cu = 'Reconfiguring' if exists(d) else 'Creating'
    L, O = '\x1b[0;38;5;240m', '\x1b[0m'
    matching_units = []

    def n(r, u=matching_units):
        n = r.name
        g = api.g
        if FLG.init_create_all_units:
            matching_units.extend(to_list(g(r, 'systemd', None)))
        # if given one by one:
        units = g(FLG, 'init_create_unit_files', [])
        cmds = api.rsc_cmds(r)
        for c in cmds:
            # that check is also in write_resource! don't change only here:
            if any([u for u in units if u == c]):
                matching_units.append(c)

        p = getattr(r, 'provides', '')
        if p:
            n += ' [%s]' % ' '.join(r.provides)
        return L + n + O if getattr(r, 'installed') else n

    rsc = '\n - '.join(n(r) for r in rscs)
    mu = list(set(matching_units))
    units = 'Unit files for:       ' + ('-' if not mu else ', '.join(mu)) + '\n'
    r = [
        '',
        cu + ' project directory %s' % d,
        '',
        'Conda resources into: %s' % api.S.conda_prefix,
        'Filesystem rscs into: %s' % api.S.fs_dir,
        units,
        'Resources: \n - %s' % rsc,
        '',
        'Confirm [Y|q|d:details|f:force non interactive install]: ',
    ]
    return '\n'.join(r)


def verify_systemctl_availability():
    if not 'linux' in sys.platform.lower():
        app.die('System platform must be Linux for unit files')
    if os.system('type systemctl'):
        app.die('Systemd must be present for unit files')


def start_editor(finder=api.find_resources_files_in_sys_path):
    m = finder().items()
    for mn, fn in m:
        print()
        app.info(fn, module=mn)
        s = read_file(fn)
        if FLG.edit_matching_resource_file in s:
            do(system, '$EDITOR "%s"' % fn)
    return mn


def delete_all_matching_service_unit_files(match):
    d = os.environ.get('HOME') + '/.config/systemd/user'
    for fn in os.listdir(d):
        if not match in fn:
            app.info('Skipping not matching unit', fn=fn)
        fn = d + '/' + fn
        if not os.path.isfile(fn):
            continue
        s = read_file(fn)
        if api.unit_match in s:
            app.warn('Unlinking unit file', fn=fn)
        app.info('Skipping unit without match string', match=api.unit_match)


class disabled:
    rscs = ()


def get_matching_resources():
    m = FLG.init_resource_match
    negates = []
    for u in list(m):
        if u.startswith('!'):
            m.remove(u)
            negates.append(u[1:])
    r = api.find_resource_defs()
    rscs = S.rscs_defined
    api.add_install_state(rscs)
    matches = lambda r: any([_ in str(api.to_dict(r)) for _ in m])
    rscs = [r for r in rscs if not m or matches(r)]
    if negates:
        n = negates
        rscs = [r for r in rscs if not any([_ in str(api.to_dict(r)) for _ in n])]
    disabled.rscs = d = [
        r for r in rscs if r.disabled == True and not matches(r) and not r.installed
    ]
    if d:
        app.info('Disabled (only via -irm)', resources=[i.name for i in d])
    [rscs.remove(r) for r in list(rscs) if r in d]
    return rscs


def run():
    if FLG.list_resources_files:
        rscs = get_matching_resources()
        app.info('Listing Defined Resources')
        app.info('details', json=rscs_dicts(rscs))
        return [r for r in rscs]
    m = FLG.delete_all_matching_service_unit_files
    if m:
        return do(delete_all_matching_service_unit_files, match=m)

    if FLG.edit_matching_resource_file:
        return start_editor()

    # the project directory:
    d = FLG.init_at
    if not d:
        app.error('No project dir given')
        return

    d = repl_dollar_var_with_env_val(d)
    d = os.path.abspath(d)
    d = d[:-1] if d.endswith('/') else d
    if exists(d):
        do(os.chdir, d)
        d = FLG.init_at = os.path.abspath('.')
        project.set_project_dir(dir=d)

    rscs = get_matching_resources()
    project.set_project_dir(dir=d)
    if FLG.init_create_unit_files:
        do(verify_systemctl_availability)

    if not (sys.stdin.isatty() and sys.stdout.isatty()) and not FLG.force:
        app.die('Require --force when run without a tty')

    if FLG.force:
        app.warn('Installing resources', resources=rscs)
    else:
        while True:
            y = input(confirm(d, rscs))
            if y.lower() in ('n', 'q'):
                app.die('Unconfirmed, stopping installation')
            if y.lower() == 'd':
                rscs_details(rscs)
                continue
            if y.lower() == 'f':
                FLG.force = True
            print()
            break

    do(create_project_dirs)
    do(os.chdir, d)

    for r in rscs:
        do(api.Install.resource, r, ll=10)

    do(git_init)
    if FLG.init_create_unit_files:
        app.info('All project file created.')
        app.info('Enabling systemd user service units.')
        if not os.system('systemctl --user --no-pager status | head -n0'):
            app.info('systemd --user available, calling daemon-reload')
            if not os.system('systemctl --user daemon-reload'):
                return app.info('Done init.')

        app.warn('systemd --user service seems not running', hint=T_SDU_SVD)
        # automatic installs must get an error when this happens:
        app.die('Failing the project init command, please fix systemd --user')


# https://serverfault.com/a/1026914 - this is really relevant,
# since RH and drives deployers crazy on first command!
T_SDU_SVD = '''
We are using the systemd **user** service to manage processes. This means there is a systemd process that runs as unprivileged user. 
The systemd user service is not used as commonly as the normal systemd process manager.
For example Red Hat disabled the systemd user service in RHEL 7 (and thereby all distros that come from RHEL, like CentOS, Oracle Linux 7, Amazon Linux 2).
However, RedHat has assured that running the systemd user service is supported as long as the service is re-enabled.

This is how to start the `systemd --user service` for user with $UID=_UID_ (the current user):

As root create this unit file:

____________________________________________________________________________________________________

# cat /etc/systemd/system/user@_UID_.service
[Unit]
Description=User Manager for UID %i
After=systemd-user-sessions.service
# These are present in the RHEL8 version of this file except that the unit is Requires, not Wants.
# It's listed as Wants here so that if this file is used in a RHEL7 settings, it will not fail.
# If a user upgrades from RHEL7 to RHEL8, this unit file will continue to work:

After=user-runtime-dir@%i.service
Wants=user-runtime-dir@%i.service

[Service]
LimitNOFILE=infinity
LimitNPROC=infinity
User=%i
PAMName=systemd-user
Type=notify
# PermissionsStartOnly is deprecated and will be removed in future versions of systemd
# This is required for all systemd versions prior to version 231
PermissionsStartOnly=true
ExecStartPre=/bin/loginctl enable-linger %i
ExecStart=-/lib/systemd/systemd --user
Slice=user-%i.slice
KillMode=mixed
Delegate=yes
TasksMax=infinity
Restart=always
RestartSec=15

[Install]
WantedBy=default.target

____________________________________________________________________________________________________

Then enable and start the unit.

Run `ps -fww $(pgrep -f "systemd --user")` to verify success, then try re-init the project.
'''

T_SDU_SVD = T_SDU_SVD.replace('_UID_', str(os.getuid()))


main = lambda: run_app(run, flags=Flags)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Running installed resources.

"""


import os, sys, json, time
from functools import partial

from devapp import app
from . import _resources_api as api
from devapp.tools import (
    FLG,
    project,
    exists,
    jsondiff,
    write_file,
    read_file,
    offset_port,
    wait_for_port,
)

app, run_app, do, system = app.app, app.run_app, app.do, app.system
from . import resources_list as rscsl
from subprocess import Popen

g = api.g
S = api.S


class Flags(rscsl.Flags):
    """Simple Process Start/Stop, Intended for Dev and Py-Testing"""

    autoshort = ''

    class dir:
        n = 'Run command after changing to this directory'

    class cmd:
        n = 'Command. Must be last argument, argv after -- will be added. You can also encluse a full command into a string, e.g. "npm -h >/dev/null".'

    class background:
        s = 'bg'
        n = 'Run in background'
        d = False

    class await_port:
        n = 'Block until port is reachable, when run in background'

    class await_port_timeout:
        d = 10

    class show:
        n = 'Print out, not run the full command'
        d = False

    class kill:
        n = 'Kill matching running resource.'
        d = False

    class killall:
        s = 'ka'
        d = False

    class shell:
        n = 'Start a shell instead of running the command - but with the environment set.'
        d = False


# --------------------------------------------------------------------------- app

fn_run = lambda: project.root() + '/.running_resources.json'


def check_running():
    S.fn_run = fn = fn_run()
    r = json.loads(read_file(fn, dflt='{}'))
    S.running_at_start = r = {k: v for k, v in r.items() if alive(v)}
    write_file(fn, json.dumps(r, indent=4))


def alive(p):
    try:
        return not os.kill(p['pid'], 0)
    except:
        return False


def kill_resource():
    fn, r = S.fn_run, S.running_at_start
    for k, v in r.items():
        if not FLG.cmd in k and not FLG.killall:
            continue
        app.warn('Killing resource', cmd=k, **v)
        os.kill(v['pid'], 15)


def find_resource_by_cmd(cmd):
    a, args = sys.argv, []
    orig_cmd = cmd
    if '--' in sys.argv:
        args = a[a.index('--') + 1 :]
    cmd += ' ' + ' '.join(args)
    scmd, args = cmd.split(' ', 1)
    rsc = api.matching_resource(scmd)

    have = [rsc.cmd] + g(rsc, 'provides', [])
    sel = [c for c in have if scmd in c]
    if len(sel) > 1:
        app.info('more matches, trying exact match', matches=sel)
        sel = [c for c in sel if scmd == c]
    if not sel or len(sel) > 1:
        app.die('No match found', have=have)
    sel = sel[0]
    breakpoint()  # FIXME BREAKPOINT
    cmd = api.Run.find_resource_cmd(rsc, sel)
    breakpoint()  # FIXME BREAKPOINT
    return cmd


def create_run_file_entry(cmd, rsc, pid):
    fn, r = S.fn_run, S.running_at_start
    r[cmd] = {
        'pid': pid,
        'rsc': api.to_dict(rsc),
        'at': time.ctime(),
        'ts': time.time(),
    }
    write_file(fn, json.dumps(r, indent=4, sort_keys=True))


def run_foregrounded(cmd, rsc):
    do(create_run_file_entry, cmd=cmd, rsc=rsc, pid=os.getpid())
    return do(system, cmd)


def run_backgrounded(cmd, rsc):
    app.info('Backgrounding', cmd=cmd)
    fnull = open(os.devnull, 'w')
    p = Popen(cmd, stderr=fnull, stdout=fnull, shell=True, close_fds=True)
    do(create_run_file_entry, cmd=cmd, rsc=rsc, pid=p.pid)
    port = FLG.await_port or g(rsc, 'wait_port') or g(rsc, 'port')
    if port:
        port = offset_port(port)
        to = max(FLG.await_port_timeout, g(rsc, 'port_wait_timeout', 0))
        wait_for_port(port, timeout=to)
        app.info('Port open', port=port, cmd=cmd)
        time.sleep(0.2)


def run():
    if FLG.dir:
        os.chdir(FLG.dir)
    cmd = FLG.cmd
    check_running()

    if not cmd and not FLG.killall:
        return S.running_at_start
    if offset_port(1) != 1:
        app.warn('GLOBAL PORT OFFSET', offset=offset_port(0))
    if FLG.kill or FLG.killall:
        return kill_resource()

    cmd, rsc = do(find_resource_by_cmd, cmd=cmd)
    if cmd in S.running_at_start:
        return app.warn('Already running', cmd=cmd, **S.running_at_start[cmd])

    if FLG.show:
        return cmd
    if FLG.shell:
        if FLG.background:
            app.die('Cannot run a shell backgrounded')
        print('Running bash instead of:\n"%s"' % cmd)
        return os.system('bash')

    return do(run_backgrounded if FLG.background else run_foregrounded, cmd, rsc)


main = partial(run_app, run, flags=Flags)



import re

# for link rewrites:
href_slash_start_single = re.compile(" href *= *'/", re.IGNORECASE)
href_slash_start_dbl = re.compile(' href *= *"/', re.IGNORECASE)
src_slash_start_single = re.compile(" src *= *'/", re.IGNORECASE)
src_slash_start_dbl = re.compile(' src *= *"/', re.IGNORECASE)


def rewrite_links(url, src, rewrites={}):
    """ fix absolute and relative links """
    prot, empty, host, path = url.split('/', 3)
    host = prot + '//' + host
    path = '/' + path
    if not path.endswith('/'):
        path = path + '/'

    # absolut links: insert the host
    # (href = "/..." -> href = "<host>/..."
    src = re.sub(href_slash_start_single, " href = '%s/" % host, src)
    src = re.sub(src_slash_start_single, " src = '%s/" % host, src)
    src = re.sub(href_slash_start_dbl, ' href = "%s/' % host, src)
    src = re.sub(src_slash_start_dbl, ' src = "%s/' % host, src)

    # fix relative links:
    # (href = "../../ -> go up the uri)
    uriparts = path.split('/')

    src = src.replace("'./", "'" + host + path)
    src = src.replace('"./', '"' + host + path)
    # do the '../../....' links:
    for hir in range(len(uriparts) - 2, 0, -1):
        rel = '../' * hir
        urip = '/'.join(uriparts[: -(hir + 1)]) + '/'
        src = src.replace("'%s" % rel, "'" + host + urip)
        src = src.replace('"%s' % rel, '"' + host + urip)

    for k, v in list(rewrites.items()):
        src = src.replace(k, v)

    return src


RE_SCRIPT_START = re.compile(r'<script', re.IGNORECASE)
RE_SCRIPT_END = re.compile(r'</script', re.IGNORECASE)


def neutralize_script_tags(src):
    res = re.sub(RE_SCRIPT_START, '<!-- SCRIPT NEUTRALIZED BY DEVAPPS', src)
    res = re.sub(RE_SCRIPT_END, '-- END SCRIPT NEUTRALIZED BY DEVAPPS', res)
    return res


if __name__ == '__main__':
    src = '''
<a href = "/foo.html">
<a href="/nospace.html">
<a href = '/foo.html'>
<a href='/nospace.html'>
<a src = "/foo.html">
<a src ='/foo.html'>

<a src = './foo1.html'>
<a src = "./foo1.html">

<a src = '../foo2.html'>
<a src = "../foo2.html">

<a src = '../../foo3.html'>
<a src = "../../foo3.html">

<a src = '../../../foo4.html'>
<a src = "../../../foo4.html">

<scripT attrs>evil stuff</sCriPt>
    '''
    src = neutralize_script_tags(src)
    res = rewrite_links('http://my.host/l1/l2/l3', src)

    assert '''<a href = "http://my.host/foo.html">''' in res
    assert '''<a href = "http://my.host/nospace.html">''' in res
    assert '''<a href = 'http://my.host/foo.html'>''' in res
    assert '''<a href = 'http://my.host/nospace.html'>''' in res
    assert '''<a src = "http://my.host/foo.html">''' in res
    assert '''<a src = 'http://my.host/foo.html'>''' in res

    assert '''<a src = 'http://my.host/l1/l2/l3/foo1.html'>''' in res
    assert '''<a src = "http://my.host/l1/l2/l3/foo1.html">''' in res

    assert '''<a src = 'http://my.host/l1/l2/foo2.html'>''' in res
    assert '''<a src = "http://my.host/l1/l2/foo2.html">''' in res

    assert '''<a src = 'http://my.host/l1/foo3.html'>''' in res
    assert '''<a src = "http://my.host/l1/foo3.html">''' in res

    assert '''<a src = 'http://my.host/foo4.html'>''' in res
    assert '''<a src = "http://my.host/foo4.html">''' in res

    assert (
        '''<!-- SCRIPT NEUTRALIZED BY DEVAPPS attrs>evil stuff-- END SCRIPT NEUTRALIZED BY DEVAPPS>'''
        in res
    )

    print('%s\nrewritten to\n%s' % (src, res))
    print(res)



# this is a marker for .pop(), otherwise we are not able to detect if a
# object is already in or not
_marker = object()

# Try to import the C-Extensions for AXTree
from axlib.ax_tree._ax_tree import _AXTree


def _build_axtree(_base_name, _base_parent_type):
    """
    Build a AXTree Class.
    Again it is possible to provide different base classes
    """

    # We need to override *all* methods from dict, otherwise
    # __getitem__, __setitem__ from _base_parent_type are NOT used.
    # They are responsible for breaking the 'dotted' keys into the tree.

    def __delitem__(self, key):
        parts = key.split('.')
        cur = self
        for n in parts[:-1]:
            cur = cur[n]
        _base_parent_type.__delitem__(cur, parts[-1])

    def number_of_leaves(self):
        return len(list(self.iter_leaf_keys()))

    def fromkeys(self, iterable, v=None):
        for key in iterable:
            self[key] = v

    def has_key(self, key):
        return key in self

    def pop(self, key, v=_marker):
        try:
            ret = self[key]
            del self[key]
            return ret
        except KeyError:
            if v is _marker:
                raise
            return v

    def setdefault(self, key, v=None):
        try:
            return self[key]
        except KeyError:
            self[key] = v
            return v

    def merge(self, tree):
        for key, value in tree.iter_leaf_items():
            self[key] = value

    # Zope RestrictedPython assumess structures (that are not dict or list) to
    # have __guarded_setitem__, __guarded_delitem__, __guarded_setattr__, and
    # __guarded_delattr__ attributes. For instance, classes not having these
    # attributes are not allowed to do assignments or del operations.
    # These statements will fail if the attributes are not included:
    #     x['5'] = 1
    #     del x['7']
    # Some background can be found in the RestrictedPython/Guards.py:103
    __guarded_setitem__ = _base_parent_type.__setitem__
    __guarded_delitem__ = _base_parent_type.__delitem__

    attributes = dict(locals())
    del attributes['_base_name']
    del attributes['_base_parent_type']
    return type(_base_name, (_base_parent_type,), attributes)


# Build an AXTree-Class where a 'plain' python dictionary is used for storage
AXTree = _build_axtree('AXTree', _AXTree)



#!/usr/bin/env python
"""
Pulling Container Filesystems from Registries.

dpull: Pull oci image layers from an oci registry.
This is a working literal translation of the jq based moby-download frozen image tool.
"""
# Could be done far smaller.
from devapp import gevent_patched
import hashlib
import json
import os
import shutil
from copy import deepcopy
from functools import partial

import requests
from devapp.app import FLG, app, run_app


class Flags:
    autoshort = ''

    class dir:
        n = 'Target dir'
        d = './images'

    class repo:
        n = 'Examples: dockerio://alpine:latest (dockerio is default), axc2://alpine:latest, ax://alpine:latest, busybox (without tag it implies :latest)'
        d = 'busybox:latest'


manifestJsonEntries = []
registries = {
    'dockerio': {
        'authService': 'registry.docker.io',
        'authBase': 'https://auth.docker.io',
        'registryBase': 'https://registry-1.docker.io',
        'mani': 'application/vnd.docker.distribution.manifest',
    }
}
this_archi = 'amd64'

registries['axc2'] = d = dict(registries['dockerio'])
d['registryBase'] = 'https://axc2.mycompany.com'

registries['axint'] = d = dict(registries['dockerio'])
d['registryBase'] = 'https://artifacts-internal.mycompany.com/docker-internal'

registries['ax'] = d = dict(registries['dockerio'])
d['registryBase'] = 'https://artifacts.mycompany.com/docker-internal'
d['authBase'] = 'https://artifacts.mycompany.com/docker-internal'


def set_registry_urls_global(reg='dockerio'):
    """restricing a single dpull process to one registry, will fix on demand"""
    m = registries[reg]
    for k in m.keys():
        globals()[k] = m[k]


ctx = {'token': None, 'D': None, 'image': None}
# doNotGenerateManifestJson = False


def log(msg, **kw):
    app.log.info(msg, **kw)


def dbg(msg, **kw):
    app.log.debug(msg, **kw)


def die(msg, **kw):
    app.die(msg, **kw)


def run_in(cmd, d):
    h = os.getcwd()
    os.chdir(d)
    log('Running cmd', cmd=cmd, dir=d)
    r = not os.system(cmd)
    if not r:
        die('Failed', cmd=cmd)
    os.chdir(h)


def showj(msg, res, **kw):
    dbg(msg, res=deepcopy(res), **kw)


def G(url, headers={}):
    res = requests.get(url, headers=headers)
    if res.status_code != 200:
        die('failed', url=url, status=res.status_code, tx=res.text)
    res = res.text
    if res and res[0] in ('{', '['):
        res = json.loads(res)
        showj('Result', res=res, url=url)
    return res


def get_token(image):
    b = authBase + '/token?service='
    h = G(''.join((b, authService, '&scope=repository:', image, ':pull')))
    log('have token', token=h['token'][:10] + '...')
    ctx['token'] = h['token']


headers = lambda: {
    'Authorization': 'Bearer %(token)s' % ctx,
    'User-Agent': 'curl/7.64.0',
}


def get_mani(image, digest):
    h = headers()
    a = []
    for k in '.v2+json', '.list.v2+json', '.v1+json':
        a.append(mani + k)
    h['Accept'] = ', '.join(a)
    res = G(registryBase + '/v2/%s/manifests/%s' % (image, digest), headers=h)
    return res


def fetch_blob(digest, fn):
    h = headers()
    ctx['bdigest'] = digest
    u = registryBase + '/v2/%(image)s/blobs/%(bdigest)s' % ctx
    log('Getting blob', url=u)
    r = requests.get(u, headers=h, stream=True)
    with open(fn, 'wb') as f:
        shutil.copyfileobj(r.raw, f)
    return fn


def sha256sum(s):
    return hashlib.sha256(bytes(s, 'utf-8')).hexdigest()


def write(fn, s):
    if isinstance(s, (tuple, list, dict)):
        showj('Result', res=s)
        s = json.dumps(s, indent=4, sort_keys=True)
    with open(fn, 'wb') as fd:
        fd.write(bytes(s, 'utf8'))


def read(fn):
    with open(fn) as fd:
        s = fd.read()
        try:
            return json.loads(s)
        except:
            return s


def handle_single_manifest_v2(manifest):
    mj = manifest
    digest = mj['config']['digest']
    img_id = digest.replace('sha256:', '')
    fn_config = '%s.json' % img_id
    fetch_blob(digest, ctx['D'] + '/' + fn_config)
    layer_fs = mj['layers']
    log('Downloading', img_id=ctx['imageIdentifier'], layers=len(layer_fs))
    layer_tars = []
    layer_id = ''
    for i in layer_fs:
        lay_media_type = i['mediaType']
        lay_digest = i['digest']
        # save the previous layer's ID
        parent_id = layer_id
        # create a new fake layer ID based on this layer's digest and the previous layer's fake ID
        layer_id = sha256sum(parent_id + '\n' + lay_digest + '\n')
        # this accounts for the possibility that an image contains the same layer twice (and thus has a duplicate digest value)
        d = ctx['D'] + '/' + layer_id
        os.system('mkdir -p "%s"' % d)
        write(d + '/VERSION', '1.0\n')
        layer_tars.append(layer_id + '/layer.tar')
        if not os.path.exists(d + '/json'):
            # fmt:off
            add = {
                'id': layer_id,
                'created': '0001-01-01T00:00:00Z',
                'container_config': {
                    'Hostname'     : '',
                    'Domainname'   : '',
                    'User'         : '',
                    'AttachStdin'  : False,
                    'AttachStdout' : False,
                    'AttachStderr' : False,
                    'Tty'          : False,
                    'OpenStdin'    : False,
                    'StdinOnce'    : False,
                    'Env'          : None,
                    'Cmd'          : None,
                    'Image'        : '',
                    'Volumes'      : None,
                    'WorkingDir'   : '',
                    'Entrypoint'   : None,
                    'OnBuild'      : None,
                    'Labels'       : None,
                },
            }
            # fmt:on
            write(d + '/json', add)
        mt = 'application/vnd.docker.image.rootfs.diff.tar.gzip'
        if lay_media_type == mt:
            lay_tar = d + '/layer.tar'
            if os.path.exists(lay_tar):
                log('Skipping, already exists', layer=lay_tar)
                continue
            fetch_blob(lay_digest, lay_tar)

    # change "$imageId" to be the ID of the last layer we added (needed for old-style "repositories" file which is created later -- specifically for older D
    # munge the top layer image manifest to have the appropriate image configuration for older daemons
    image_id = layer_id
    fn = '/'.join((ctx['D'], image_id, 'json'))
    m = read(fn)
    img_old_cfg = {'id': m['id']}
    p = m.get('parent')
    if p:
        img_old_cfg['parent'] = p
    m = read(ctx['D'] + '/' + fn_config)
    m.pop('history', 0)
    m.pop('rootfs', 0)
    m.update(img_old_cfg)
    write(fn, m)
    manifestJsonEntries.append(
        {
            'Config': fn_config,
            'RepoTags': [ctx['image'].replace('libaray/', '') + ':' + ctx['tag']],
            'Layers': layer_tars,
        }
    )
    return image_id


def do_v1():
    M = ctx['manifestJson']
    fs_layers = [m['blobSum'] for m in M['fsLayers']]
    # imageIdentifier uses schemaVersion= %(schemaVersion)shhhhhh
    # this script cannot (currently) recreate the 'image config' to put in a
    # 'manifest.json' (thus any schemaVersion 2+ images will be imported in
    # the old way, and their 'docker history' will suffer)
    histories = M['history']
    history = [json.loads(histories[i]['v1Compatibility']) for i in range(len(histories))]
    log('History', history=history)
    image_id = history[0]['id']
    log(
        'Downloading', imageIdentifier=ctx['imageIdentifier'], layers=len(history),
    )
    nr = 0
    for h in history:
        layer_id = h['id']
        d = ctx['D'] + '/' + layer_id
        os.system('mkdir -p "%s"' % d)
        write(d + '/VERSION', '1.0\n')
        write(d + '/json', h)
        if os.path.exists(d + '/layer.tar'):
            log('Skipping, exists', layer=d)
        else:
            fetch_blob(fs_layers[nr], d + '/layer.tar')
        nr += 1
    return image_id


def do_v2():
    M = ctx['manifestJson']
    mediaType = M.get('mediaType')
    if mediaType == mani + '.v2+json':
        img_id = handle_single_manifest_v2(M)
    elif mediaType == mani + '.list.v2+json':
        manifests = M['manifests']
        layerFS = manifests
        found = []
        for l in layerFS:
            if not l['platform']['architecture'] == this_archi:
                continue
            found.append(1)
            digest = l['digest']
            sub_mani = get_mani(ctx['image'], digest)
            img_id = handle_single_manifest_v2(sub_mani)
        if not found:
            die('No manifest found for architecture', archi=this_archi)
    else:
        die('Unknown manifest media type', mediaType=mediaType)
    return img_id


def run():
    # FLG.log_dev_fmt_coljson = ['res']
    D = FLG.dir
    if not os.path.exists(D):
        os.system('mkdir -p "%s"' % D)
    run = partial(run_in, d=D)
    run('rm -f tags-*.tmp', d=D)
    repo = FLG.repo
    if not '://' in repo:
        repo = 'dockerio://' + repo
    if not len(repo.split(':')) == 3:
        repo += ':latest'
    app.info('Repo taken', repo=repo)
    reg, repo = repo.split('://')
    set_registry_urls_global(reg)
    image, imageTag = repo.split(':', 1)
    tag, digest = (imageTag + '@').split('@')[:2]
    if not digest:
        digest = tag
    # no digest yet
    oimg = image
    if not '/' in image:
        image = 'library/' + image

    imageFile = image.replace('/', '_')
    get_token(image)
    manifestJson = get_mani(image, digest)
    imageIdentifier = '%s:%s@%s' % (image, tag, digest)
    schemaVersion = manifestJson['schemaVersion']
    ctx.update(**locals())

    if schemaVersion == 2:
        img_id = do_v2()
    else:
        img_id = do_v1()

    R = ctx['D'] + '/repositories'
    r = read(R) if os.path.exists(R) else {}
    r.setdefault(oimg, {}).update({ctx['tag']: img_id})
    write(R, r)
    write(ctx['D'] + '/manifestJson', manifestJsonEntries)
    log('Success', docker_load_cmd="tar -cC '%s' . | docker load" % D)


main = lambda: run_app(
    run,
    flags=Flags,
    kw_log=dict(
        log_dev_fmt_coljson=['res'], censor=(['res', 'token'], ['res', 'access_token']),
    ),
)


if __name__ == '__main__':
    main()



import json
import os

env = os.environ
_parsed = []
parsed_links = {}


def links():
    pl = parsed_links
    if not _parsed:
        with open(os.environ['build_dir'] + '/links.json') as fd:
            j = fd.read()
        parsed_links.update(json.loads(j))
        _parsed.append(True)
    return pl


# create string fid lists from classes list at build time:
full_id = lambda cls: [c.name for c in cls._parents]
fids = lambda clses: ['.'.join(full_id(co)) for co in clses]


def _get_via(for_, links, io, *caps, min=1, max=1000, **kw):
    """
    conn = lambda io, w: api.links.get_via(r, io, w, min=1)[0]
    httpapi = conn('in', 'api')
    """

    res = []
    for l in links[io]:
        have = [c for c in caps if c in l['via']['caps']]
        if len(have) == len(caps):
            v = dict(l['via'])
            # that can be list of strings (when loaded from links.json
            # or still a list of clses (at build time)
            # we return always the strings:
            c = l['cls']
            v['cls'] = [] if not c else c if isinstance(c[0], str) else fids(c)
            res.append(v)
    if len(res) < min or len(res) > max:
        msg = '%s: have %s links for %s %s, require min %s, max %s'
        msg = msg % (for_, len(res), '+'.join(caps), io, min, max)
        raise Exception(msg)
    return res


def get_via_by_r(r, io, *caps, min=1, max=1000):
    """Api method while rendering templates, (build time)"""
    return _get_via(r['___'], r['links'], io, *caps, min=min, max=max)


def get_via(io, *caps, min=1, max=1000, **kw):
    """Api method for apps, while running (run time)"""
    l = links()
    return _get_via(env['DA_CLS'], l, io, *caps, min=min, max=max, **kw)



#!/usr/bin/env python
import os
from json import loads

from devapp.app import app
from devapp.tools import abspath, clean_env_key, dirname, exists, read_file, sp_call

env = os.environ
is_svc = lambda: env['type'] == 'Service'


def fs_build_dirs():
    m = dict([('$' + k, v % env) for k, v in d_fs])
    m['$fs'] = '' if is_svc() else m['$bd'] + '/fs'
    return m


d_fs = dict(
    DA_DIR='%(DA_DIR)s',
    bd='%(DA_DIR)s/build/%(DA_CLS)s',
    rd='%(DA_DIR_REPOS)s',
    # sd='%(DA_DIR_SNAPS)s',
).items()


def already_overlayed(frm, to):
    """ Not so easy since dirs can be shadowed, the only secure way is mounts"""
    overs = [l for l in read_file('/proc/mounts').splitlines()]
    # TODO!!
    return False


class FS:
    """
    This is in use by the app starter, before app is made
    WE HAVE NO DEVAPP.APP yet! run.py does set a fake InitApp class into us
    at process start, for basic logging.
    """

    initapp = None  # set by run.py to basic logging only mock
    inode = lambda fn: os.stat(fn).st_ino
    mkdir = lambda d: sp_call('mkdir', '-p', d) if not exists(d) else 0
    d_build = lambda e: '%(DA_DIR)s/build/%(DA_CLS)s' % e

    def app(level, msg, **kw):
        # just for logging before and after we have the app - see usage below:
        # must work with the fake init app as well, thats why:
        f = getattr(app, level, None) or getattr(FS.initapp, level)
        f(msg, **kw)

    def build_from_fs_stack(mode='shared', app_env=env, fs=None):
        supported = {'shared': ['symlink', 'copy_file', 'tarpipe']}
        run = supported.get(mode)
        if run is None:
            raise Exception('Not supported currently: FS building %s' % mode)
        # meth shared: built time
        e = app_env  # at build time r['env'], else sourced os.env
        name = clean_env_key(e['DA_CLS'])
        env_have_flag = 'da_have_%s_fs_%s' % (mode, name)
        if e.get(env_have_flag):
            return
        # runtime:
        if fs is None:
            # info('Building filesystem...')
            fs = read_file(FS.d_build(e) + '/fs_stack.json')
            # $bd and such. not in anymore:
            # for k, v in fs_build_dirs().items():
            #    fs = fs.replace(k, v)
            fs = loads(fs)

        # the sorting guarantees outer dirs set before their content inside
        # e.g. /etc mounted before /etc/systemd/system copied
        for k in sorted(fs, key=len):
            src, meth = fs[k]['from'], fs[k]['meth']
            if not meth in run:
                continue
            FS.app('info', meth, to=k, frm=src)
            getattr(FS, meth)(frm=src, to=k)
        # at build write env will happen for shared meths, at proc run we just need the export
        e[env_have_flag] = name

    def overlay(frm, to):
        """Overlay assumes to be the ONLY one done, i.e. in frm we do have
        a list of all lowers
        The hard part is to find out if the overlay already exists.
        """
        # FS.indod after overlay is frm[0] == to... don't trust that to check
        # if already done. Rather this:
        app = FS.app

        lower = '' + ':'.join(frm) + ''
        upper = to + '.upper'
        if already_overlayed(frm, upper):
            FS.app('debug', 'Already overlayed', target=to)
            return
        work = to + '.work'
        name = env.get('DA_CLS')
        for k in upper, work, to:
            if not exists(k):
                os.mkdir(k)
        overlay = [
            'mount',
            '-t',
            'overlay',
            name,
            '-o',
            'lowerdir=%s,upperdir=%s,workdir=%s' % (lower, upper, work),
            to,
        ]
        return FS.sp_call(*overlay)

    def can_mount(c=[]):
        """Non root requires fusefs/bindfs"""
        if not c:
            c.append(
                FS.bindmount
                if not os.geteuid()
                else FS.bindfsmount
                if os.system('hash bindfs 2>/dev/null') == 0
                else False
            )
        return c[0]

    def mount(frm, to):
        meth = FS.can_mount()
        app = FS.app
        if not meth:
            m = 'Cannot perform mount as user w/o bindfs'
            app('die', m, frm=frm, dest=to, remedy='Install bindfs')

        if os.path.isdir(frm):
            FS.mkdir(to)
        else:
            FS.mkdir(dirname(to))
            FS.sp_call('touch', to)
        if FS.inode(frm) == FS.inode(to):
            # inode check works also with bindfs
            FS.app('debug', 'Already mounted', frm=frm)
            return
        return meth(frm, to)

    def bindmount(frm, to):
        return FS.sp_call('mount', '--bind', frm, to)

    def bindfsmount(frm, to):
        # not in use. bubblewrap better?
        return FS.sp_call('bindfs', '-o', 'nonempty', '-n', frm, to)

    def copy_file(frm, to):
        os.unlink(to) if exists(to) else 0
        return FS.sp_call('cp', frm, to)

    def tarpipe(frm, to, create=False):
        if not exists(to):
            if not create:
                raise Exception('Missing: %s' % to)
            FS.sp_call('mkdir -p "%s"' % to)
        cmd = '(cd "%s" && tar -cf - .) | (cd "%s" && tar -xpf -)'
        return FS.sp_call(cmd % (frm, to), name='tarpipe')

    def symlink(frm, to):
        if not exists(frm):
            raise Exception('Symlink source not exists %s' % frm)
        app = FS.app
        if os.path.islink(to):
            cur = os.readlink(to)
            if cur == frm:
                return
            os.unlink(to)
        if exists(to):
            raise Exception(
                'Symlink Target %s exists and is no symlink: %s' % (frm, to)
            )
        FS.mkdir(dirname(to))
        return FS.sp_call('ln', '-s', frm, to)

    def sp_call(*args, name=None):
        name, params = (name, args) if name else (args[0], args[1:])
        FS.app('debug', name, args=params)
        if len(args) == 1:
            err = os.system(args[0])
        else:
            err = sp_call(*args)
            if err:
                raise Exception('Error', err)


def full_url_and_path(env, url, mode):
    orig_url = url
    if url.startswith('http') or url.startswith('file://') or exists(url):
        if exists(url):
            url = 'file://' + abspath(url)
        return url, url[8:].split('/', 1)[1]
    if mode == 'oci':
        parts = url.split(':')
        if len(parts) == 1:
            app.die('No oci image format', url=url)
        if len(parts) == 2:
            # can't express auth and images endpoint in a single url, need
            # that indirection:
            return 'dockerio://%s' % url, url.replace(':', '')
        else:
            return url, url.split(':', 1)[1].replace(':', '')
    elif mode == 'git':
        if 'git@' in url:
            if ':' in url:
                return url, url.rsplit(':')[-1]
            return url, url.split('/', 1)[1]
        elif url.startswith('git://'):
            return url, url[8:].split('/', 1)[1]
    else:
        app.die('Not supported', mode=mode, url=url)
    if url[0] in ('/', ':'):
        url = url[1:]
    if not ':' in url:
        url = ':' + url
    pres, url = url.split(':', 1)
    if not pres:
        pre = 'DA_URL_%s' % mode.upper()
    else:
        pre = 'DA_URL_%s' % pres
    if not pre in env:
        app.die(
            'URL shortcut specified but not found in environ',
            url=orig_url,
            missing=pre,
            short=pres,
        )
    furl = env[pre] + '%s' + url
    sep = ':' if 'git@' in furl else '/'
    return furl % sep, url



#!/usr/bin/env python
import json
import os
import sys
from configparser import ConfigParser
from functools import partial

import requests
from devapp.app import FLG, app
from devapp.cli_flags import oauth
from devapp.spec.os_tools import confirm
from devapp.spec.os_tools import env_get_interactive_write as envget
from theming.colorhilite import coljhighlight

get_endpoint = oauth()

env = os.environ
exists = os.path.exists
fn_cfg = lambda: env['DA_DIR'] + '/secure/python-gitlab.cfg'


def cfg(c=[], all=None):
    """returns parsed config for section <endpoint_name> - if not exists, builds it
    When interactive auth is requred we do not write - i.e. only valid for cur.
    session.
    To avoid frequent password queries, call GL.setup function.
    """
    try:
        return c[0][FLG.endpoint_name]
    except:
        pass
    c.append(ConfigParser())
    fn = fn_cfg()
    if exists(fn):
        app.info('Reading', fn=fn)
        c[0].read(fn)
    else:
        c[0]['global'] = {
            'default': FLG.endpoint_name,
            'ssl_verify': 'true',
            'timeout': 1,
        }
    if not FLG.endpoint_name in c[0]:
        t = GitLab.authenticate()
        c[0][FLG.endpoint_name] = t
    if all:
        return c[0]
    return c[0][FLG.endpoint_name]


def req(meth, pth, **kw):
    t = cfg()
    u = t['url'] + '/api/v4' + pth
    kw['access_token'] = t['access_token']
    app.info('API request', url=u, **kw)
    d = kw.get('data')
    m = partial(meth, json=d) if d else partial(meth, params=kw)
    res = m(u, headers={'ContentType': 'application/json'})
    if res.status_code < 300:
        try:
            return json.loads(res.text)
        except:
            return {'status': res.status_code, 'response': res.text}
    app.die('Error', status_code=res.status_code)


def get(pth, **kw):
    return req(requests.get, pth, **kw)


def delete(pth, **kw):
    return req(requests.delete, pth, **kw)


def post(pth, data):
    return req(requests.post, pth, data=data)


class GitLab:
    """Interacting with the Gitlab Server"""

    def authenticate(user=None, password=None, gl_endpoint=None):
        """User Password Auth to get oauth2 token"""
        user = user or FLG.username or envget('user')
        password = password or envget('password')
        h = gl_endpoint or get_endpoint()
        app.info('Getting tokens', endpoint=h, user=user, pw=password[:2] + '..')
        d = {'grant_type': 'password', 'username': user, 'password': password}
        r = requests.post(h + '/oauth/token', json=d)
        if not r.status_code == 200:
            app.die('Sorry', status=r.status_code)
        t = json.loads(r.text)
        t['url'] = h
        return t

    def setup():
        """write oauth token from authenticate into config file"""
        c = cfg(all=True)
        with open(fn_cfg(), 'w') as fd:
            c.write(fd)
        return app.info('Written', **cfg())

    class Projects:
        pth = '/projects'

        @classmethod
        def list(cls, id=None, search=None, owned=True, simple=True, **filters):
            """
            - gitl proj    # all
            - gitl proj 68 # details
            - gitl proj wifi # match
            """
            if str(id).isdigit():
                return cls.get(id)
            if id and not search:
                search = id

            l, kw = locals(), filters
            if search:
                kw['search'] = search
            for k in 'owned', 'simple':
                kw[k] = l[k]
            return get('/projects', **kw)

        def get(id):
            return get('/projects/%s' % id)

        default = list

    class Runners:
        @classmethod
        def add(
            cls,
            token,
            tag_list,
            description='',
            active=True,
            locked=True,
            run_untagged=False,
            protected=True,
        ):
            """
            token from <repourl>/settings/ci_cd
            - gi ru add xxasdfauYsaz-iuzo_K8G 'foo,bar'
            """
            kw = {'token': token, 'tag_list': tag_list}
            res = post('/runners', data=kw)
            return cls.get(res['id'])

        @classmethod
        def list(cls, id=None, type='a', **filters):
            """
            types: p -> project_type, i->instance_type, g->group_type
            - gi run li
            - gi run li p  # project type only
            - gi run li 42 # details
            """
            if str(id).isdigit():
                return cls.get(id)

            f = filters
            typs = {
                'p': 'project_type',
                'i': 'instance_type',
                'g': 'group_type',
                'a': 'all',
            }
            if id in typs:
                type = id
            t = typs.get(type, type)
            if t == 'all':
                r = {}
                for k in typs:
                    if k != 'a':
                        r[typs[k]] = cls.list(type=k, **filters)
                return r
            f['type'] = t
            all = filters.get('all', '')
            all = '/all' if all else ''
            return get('/runners' + all, **f)

        def all(type='i', **filters):
            """all: List all runners (requires admin)"""
            filters['all'] = True
            return GitLab.Runners.list(type=type, **filters)

        def get(id):
            return get('/runners/%s' % id)

        @classmethod
        def delete(cls, id, force=False):
            if sys.stdin.isatty() and not force:
                res = cls.get(id)
                if not res:
                    app.die('No runner found', id=id)
                print(coljhighlight(res))
                if not confirm('Delete', False):
                    return
            elif not force:
                app.die('Set to true when run non interactively')
            return delete('/runners/%s' % id)

        default = list



"""
Hilites serializable structers in the terminal.

json, yml for now

"""
# TODO: perf improvement for python objects: the serialization / parsing in again
# by pygments can be avoided by tokenizing directly based on the nature of the
# objects.
# But: This is anyway mainly for foreground, i.e. debugging...

from json import dumps

from ax.utils.six import StringIO
from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers import JsonLexer, YamlLexer
from yaml import dump, safe_dump

ysl = ytermf = jsl = termf = ''


def coljhighlight(s):
    global jsl, termf
    if not jsl:
        jsl = JsonLexer()
        termf = TerminalFormatter(bg='dark')
    io = StringIO()
    if not isinstance(s, basestring):
        s = dumps(s, indent=4, sort_keys=1, default=str)
    highlight(s, jsl, termf, io)
    res = io.getvalue()
    io.close()
    return res.replace('\\n', '\n')


def colyhighlight(s):
    global ysl, ytermf
    if not ysl:
        ysl = YamlLexer()
        ytermf = TerminalFormatter(bg='dark')
    io = StringIO()
    if not isinstance(s, basestring):
        try:
            s = safe_dump(s)
        except:
            s = dump(s, default_flow_style=False)
    highlight(s, ysl, ytermf, io)
    res = io.getvalue()
    io.close()
    return res



#!/usr/bin/env python
#
# Copyright (c) 2009-2013, Luke Maurits <luke@maurits.id.au>
# All rights reserved.
# With contributions from:
#  * Chris Clark
#  * Klein Stephane
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__version__ = '0.7.2'

import copy
import csv
import random
import re
import sys
import textwrap
import unicodedata

# from ax.utils.six.moves import zip, map

py3k = sys.version_info[0] >= 3
if py3k:
    str = str
    str = str
    uni_chr = chr
    from html.parser import HTMLParser
else:
    uni_chr = chr
    from html.parser import HTMLParser

if py3k and sys.version_info[1] >= 2:
    from html import escape
else:
    from cgi import escape

# hrule styles
FRAME = 0
ALL = 1
NONE = 2
HEADER = 3

# Table styles
DEFAULT = 10
MSWORD_FRIENDLY = 11
PLAIN_COLUMNS = 12
RANDOM = 20

_re = re.compile('\\033\\[[0-9;]*m')


def _get_size(text):
    lines = text.split('\n')
    height = len(lines)
    width = max([_str_block_width(line) for line in lines])
    return (width, height)


class PrettyTable(object):
    def __init__(self, field_names=None, **kwargs):

        """Return a new PrettyTable instance

        Arguments:

        encoding - Unicode encoding scheme used to decode any encoded input
        field_names - list or tuple of field names
        fields - list or tuple of field names to include in displays
        start - index of first data row to include in output
        end - index of last data row to include in output PLUS ONE (list slice style)
        header - print a header showing field names (True or False)
        header_style - stylisation to apply to field names in header ("cap", "title", "upper", "lower" or None)
        border - print a border around the table (True or False)
        hrules - controls printing of horizontal rules after rows.  Allowed values: FRAME, HEADER, ALL, NONE
        vrules - controls printing of vertical rules between columns.  Allowed values: FRAME, ALL, NONE
        int_format - controls formatting of integer data
        float_format - controls formatting of floating point data
        padding_width - number of spaces on either side of column data (only used if left and right paddings are None)
        left_padding_width - number of spaces on left hand side of column data
        right_padding_width - number of spaces on right hand side of column data
        vertical_char - single character string used to draw vertical lines
        horizontal_char - single character string used to draw horizontal lines
        junction_char - single character string used to draw line junctions
        sortby - name of field to sort rows by
        sort_key - sorting key function, applied to data points before sorting
        valign - default valign for each row (None, "t", "m" or "b")
        reversesort - True or False to sort in descending or ascending order"""

        self.encoding = kwargs.get('encoding', 'UTF-8')

        # Data
        self._field_names = []
        self._align = {}
        self._valign = {}
        self._max_width = {}
        self._rows = []
        if field_names:
            self.field_names = field_names
        else:
            self._widths = []

        # Options
        self._options = 'start end fields header border sortby reversesort sort_key attributes format hrules vrules'.split()
        self._options.extend(
            'int_format float_format padding_width left_padding_width right_padding_width'.split()
        )
        self._options.extend(
            'vertical_char horizontal_char junction_char header_style valign xhtml print_empty'.split()
        )
        for option in self._options:
            if option in kwargs:
                self._validate_option(option, kwargs[option])
            else:
                kwargs[option] = None

        self._start = kwargs['start'] or 0
        self._end = kwargs['end'] or None
        self._fields = kwargs['fields'] or None

        if kwargs['header'] in (True, False):
            self._header = kwargs['header']
        else:
            self._header = True
        self._header_style = kwargs['header_style'] or None
        if kwargs['border'] in (True, False):
            self._border = kwargs['border']
        else:
            self._border = True
        self._hrules = kwargs['hrules'] or FRAME
        self._vrules = kwargs['vrules'] or ALL

        self._sortby = kwargs['sortby'] or None
        if kwargs['reversesort'] in (True, False):
            self._reversesort = kwargs['reversesort']
        else:
            self._reversesort = False
        self._sort_key = kwargs['sort_key'] or (lambda x: x)

        self._int_format = kwargs['int_format'] or {}
        self._float_format = kwargs['float_format'] or {}
        self._padding_width = kwargs['padding_width'] or 1
        self._left_padding_width = kwargs['left_padding_width'] or None
        self._right_padding_width = kwargs['right_padding_width'] or None
        self._vertical_char = kwargs['vertical_char'] or self._unicode('|')
        self._horizontal_char = kwargs['horizontal_char'] or self._unicode('-')
        self._junction_char = kwargs['junction_char'] or self._unicode('+')

        if kwargs['print_empty'] in (True, False):
            self._print_empty = kwargs['print_empty']
        else:
            self._print_empty = True
        self._format = kwargs['format'] or False
        self._xhtml = kwargs['xhtml'] or False
        self._attributes = kwargs['attributes'] or {}

    def _unicode(self, value):
        if not isinstance(value, str):
            value = str(value)
        if not isinstance(value, str):
            value = str(value, self.encoding, 'strict')
        return value

    def _justify(self, text, width, align):
        excess = width - _str_block_width(text)
        if align == 'l':
            return text + excess * ' '
        elif align == 'r':
            return excess * ' ' + text
        else:
            if excess % 2:
                # Uneven padding
                # Put more space on right if text is of odd length...
                if _str_block_width(text) % 2:
                    return (excess // 2) * ' ' + text + (excess // 2 + 1) * ' '
                # and more space on left if text is of even length
                else:
                    return (excess // 2 + 1) * ' ' + text + (excess // 2) * ' '
                # Why distribute extra space this way?  To match the behaviour of
                # the inbuilt str.center() method.
            else:
                # Equal padding on either side
                return (excess // 2) * ' ' + text + (excess // 2) * ' '

    def __getattr__(self, name):

        if name == 'rowcount':
            return len(self._rows)
        elif name == 'colcount':
            if self._field_names:
                return len(self._field_names)
            elif self._rows:
                return len(self._rows[0])
            else:
                return 0
        else:
            raise AttributeError(name)

    def __getitem__(self, index):

        new = PrettyTable()
        new.field_names = self.field_names
        for attr in self._options:
            setattr(new, '_' + attr, getattr(self, '_' + attr))
        setattr(new, '_align', getattr(self, '_align'))
        if isinstance(index, slice):
            for row in self._rows[index]:
                new.add_row(row)
        elif isinstance(index, int):
            new.add_row(self._rows[index])
        else:
            raise Exception(
                'Index %s is invalid, must be an integer or slice' % str(index)
            )
        return new

    if py3k:

        def __str__(self):
            return self.__unicode__()

    else:

        def __str__(self):
            return self.__unicode__().encode(self.encoding)

    def __unicode__(self):
        return self.get_string()

    ##############################
    # ATTRIBUTE VALIDATORS       #
    ##############################

    # The method _validate_option is all that should be used elsewhere in the code base to validate options.
    # It will call the appropriate validation method for that option.  The individual validation methods should
    # never need to be called directly (although nothing bad will happen if they *are*).
    # Validation happens in TWO places.
    # Firstly, in the property setters defined in the ATTRIBUTE MANAGMENT section.
    # Secondly, in the _get_options method, where keyword arguments are mixed with persistent settings

    def _validate_option(self, option, val):
        if option in ('field_names'):
            self._validate_field_names(val)
        elif option in (
            'start',
            'end',
            'max_width',
            'padding_width',
            'left_padding_width',
            'right_padding_width',
            'format',
        ):
            self._validate_nonnegative_int(option, val)
        elif option in ('sortby'):
            self._validate_field_name(option, val)
        elif option in ('sort_key'):
            self._validate_function(option, val)
        elif option in ('hrules'):
            self._validate_hrules(option, val)
        elif option in ('vrules'):
            self._validate_vrules(option, val)
        elif option in ('fields'):
            self._validate_all_field_names(option, val)
        elif option in ('header', 'border', 'reversesort', 'xhtml', 'print_empty'):
            self._validate_true_or_false(option, val)
        elif option in ('header_style'):
            self._validate_header_style(val)
        elif option in ('int_format'):
            self._validate_int_format(option, val)
        elif option in ('float_format'):
            self._validate_float_format(option, val)
        elif option in ('vertical_char', 'horizontal_char', 'junction_char'):
            self._validate_single_char(option, val)
        elif option in ('attributes'):
            self._validate_attributes(option, val)
        else:
            raise Exception('Unrecognised option: %s!' % option)

    def _validate_field_names(self, val):
        # Check for appropriate length
        if self._field_names:
            try:
                assert len(val) == len(self._field_names)
            except AssertionError:
                raise Exception(
                    'Field name list has incorrect number of values, (actual) %d!=%d (expected)'
                    % (len(val), len(self._field_names))
                )
        if self._rows:
            try:
                assert len(val) == len(self._rows[0])
            except AssertionError:
                raise Exception(
                    'Field name list has incorrect number of values, (actual) %d!=%d (expected)'
                    % (len(val), len(self._rows[0]))
                )
        # Check for uniqueness
        try:
            assert len(val) == len(set(val))
        except AssertionError:
            raise Exception('Field names must be unique!')

    def _validate_header_style(self, val):
        try:
            assert val in ('cap', 'title', 'upper', 'lower', None)
        except AssertionError:
            raise Exception(
                'Invalid header style, use cap, title, upper, lower or None!'
            )

    def _validate_align(self, val):
        try:
            assert val in ['l', 'c', 'r']
        except AssertionError:
            raise Exception('Alignment %s is invalid, use l, c or r!' % val)

    def _validate_valign(self, val):
        try:
            assert val in ['t', 'm', 'b', None]
        except AssertionError:
            raise Exception('Alignment %s is invalid, use t, m, b or None!' % val)

    def _validate_nonnegative_int(self, name, val):
        try:
            assert int(val) >= 0
        except AssertionError:
            raise Exception('Invalid value for %s: %s!' % (name, self._unicode(val)))

    def _validate_true_or_false(self, name, val):
        try:
            assert val in (True, False)
        except AssertionError:
            raise Exception('Invalid value for %s!  Must be True or False.' % name)

    def _validate_int_format(self, name, val):
        if val == '':
            return
        try:
            assert type(val) in (str, str)
            assert val.isdigit()
        except AssertionError:
            raise Exception(
                'Invalid value for %s!  Must be an integer format string.' % name
            )

    def _validate_float_format(self, name, val):
        if val == '':
            return
        try:
            assert type(val) in (str, str)
            assert '.' in val
            bits = val.split('.')
            assert len(bits) <= 2
            assert bits[0] == '' or bits[0].isdigit()
            assert bits[1] == '' or bits[1].isdigit()
        except AssertionError:
            raise Exception(
                'Invalid value for %s!  Must be a float format string.' % name
            )

    def _validate_function(self, name, val):
        try:
            assert hasattr(val, '__call__')
        except AssertionError:
            raise Exception('Invalid value for %s!  Must be a function.' % name)

    def _validate_hrules(self, name, val):
        try:
            assert val in (ALL, FRAME, HEADER, NONE)
        except AssertionError:
            raise Exception(
                'Invalid value for %s!  Must be ALL, FRAME, HEADER or NONE.' % name
            )

    def _validate_vrules(self, name, val):
        try:
            assert val in (ALL, FRAME, NONE)
        except AssertionError:
            raise Exception(
                'Invalid value for %s!  Must be ALL, FRAME, or NONE.' % name
            )

    def _validate_field_name(self, name, val):
        try:
            assert (val in self._field_names) or (val is None)
        except AssertionError:
            raise Exception('Invalid field name: %s!' % val)

    def _validate_all_field_names(self, name, val):
        try:
            for x in val:
                self._validate_field_name(name, x)
        except AssertionError:
            raise Exception('fields must be a sequence of field names!')

    def _validate_single_char(self, name, val):
        return
        try:
            assert _str_block_width(val) == 1
        except AssertionError:
            raise Exception(
                'Invalid value for %s!  Must be a string of length 1.' % name
            )

    def _validate_attributes(self, name, val):
        try:
            assert isinstance(val, dict)
        except AssertionError:
            raise Exception('attributes must be a dictionary of name/value pairs!')

    ##############################
    # ATTRIBUTE MANAGEMENT       #
    ##############################

    def _get_field_names(self):
        return self._field_names
        """The names of the fields

        Arguments:

        fields - list or tuple of field names"""

    def _set_field_names(self, val):
        val = [self._unicode(x) for x in val]
        self._validate_option('field_names', val)
        if self._field_names:
            old_names = self._field_names[:]
        self._field_names = val
        if self._align and old_names:
            for old_name, new_name in zip(old_names, val):
                self._align[new_name] = self._align[old_name]
            for old_name in old_names:
                if old_name not in self._align:
                    self._align.pop(old_name)
        else:
            for field in self._field_names:
                self._align[field] = 'c'
        if self._valign and old_names:
            for old_name, new_name in zip(old_names, val):
                self._valign[new_name] = self._valign[old_name]
            for old_name in old_names:
                if old_name not in self._valign:
                    self._valign.pop(old_name)
        else:
            for field in self._field_names:
                self._valign[field] = 't'

    field_names = property(_get_field_names, _set_field_names)

    def _get_align(self):
        return self._align

    def _set_align(self, val):
        self._validate_align(val)
        for field in self._field_names:
            self._align[field] = val

    align = property(_get_align, _set_align)

    def _get_valign(self):
        return self._valign

    def _set_valign(self, val):
        self._validate_valign(val)
        for field in self._field_names:
            self._valign[field] = val

    valign = property(_get_valign, _set_valign)

    def _get_max_width(self):
        return self._max_width

    def _set_max_width(self, val):
        self._validate_option('max_width', val)
        for field in self._field_names:
            self._max_width[field] = val

    max_width = property(_get_max_width, _set_max_width)

    def _get_fields(self):
        """List or tuple of field names to include in displays

        Arguments:

        fields - list or tuple of field names to include in displays"""
        return self._fields

    def _set_fields(self, val):
        self._validate_option('fields', val)
        self._fields = val

    fields = property(_get_fields, _set_fields)

    def _get_start(self):
        """Start index of the range of rows to print

        Arguments:

        start - index of first data row to include in output"""
        return self._start

    def _set_start(self, val):
        self._validate_option('start', val)
        self._start = val

    start = property(_get_start, _set_start)

    def _get_end(self):
        """End index of the range of rows to print

        Arguments:

        end - index of last data row to include in output PLUS ONE (list slice style)"""
        return self._end

    def _set_end(self, val):
        self._validate_option('end', val)
        self._end = val

    end = property(_get_end, _set_end)

    def _get_sortby(self):
        """Name of field by which to sort rows

        Arguments:

        sortby - field name to sort by"""
        return self._sortby

    def _set_sortby(self, val):
        self._validate_option('sortby', val)
        self._sortby = val

    sortby = property(_get_sortby, _set_sortby)

    def _get_reversesort(self):
        """Controls direction of sorting (ascending vs descending)

        Arguments:

        reveresort - set to True to sort by descending order, or False to sort by ascending order"""
        return self._reversesort

    def _set_reversesort(self, val):
        self._validate_option('reversesort', val)
        self._reversesort = val

    reversesort = property(_get_reversesort, _set_reversesort)

    def _get_sort_key(self):
        """Sorting key function, applied to data points before sorting

        Arguments:

        sort_key - a function which takes one argument and returns something to be sorted"""
        return self._sort_key

    def _set_sort_key(self, val):
        self._validate_option('sort_key', val)
        self._sort_key = val

    sort_key = property(_get_sort_key, _set_sort_key)

    def _get_header(self):
        """Controls printing of table header with field names

        Arguments:

        header - print a header showing field names (True or False)"""
        return self._header

    def _set_header(self, val):
        self._validate_option('header', val)
        self._header = val

    header = property(_get_header, _set_header)

    def _get_header_style(self):
        """Controls stylisation applied to field names in header

        Arguments:

        header_style - stylisation to apply to field names in header ("cap", "title", "upper", "lower" or None)"""
        return self._header_style

    def _set_header_style(self, val):
        self._validate_header_style(val)
        self._header_style = val

    header_style = property(_get_header_style, _set_header_style)

    def _get_border(self):
        """Controls printing of border around table

        Arguments:

        border - print a border around the table (True or False)"""
        return self._border

    def _set_border(self, val):
        self._validate_option('border', val)
        self._border = val

    border = property(_get_border, _set_border)

    def _get_hrules(self):
        """Controls printing of horizontal rules after rows

        Arguments:

        hrules - horizontal rules style.  Allowed values: FRAME, ALL, HEADER, NONE"""
        return self._hrules

    def _set_hrules(self, val):
        self._validate_option('hrules', val)
        self._hrules = val

    hrules = property(_get_hrules, _set_hrules)

    def _get_vrules(self):
        """Controls printing of vertical rules between columns

        Arguments:

        vrules - vertical rules style.  Allowed values: FRAME, ALL, NONE"""
        return self._vrules

    def _set_vrules(self, val):
        self._validate_option('vrules', val)
        self._vrules = val

    vrules = property(_get_vrules, _set_vrules)

    def _get_int_format(self):
        """Controls formatting of integer data
        Arguments:

        int_format - integer format string"""
        return self._int_format

    def _set_int_format(self, val):
        #        self._validate_option("int_format", val)
        for field in self._field_names:
            self._int_format[field] = val

    int_format = property(_get_int_format, _set_int_format)

    def _get_float_format(self):
        """Controls formatting of floating point data
        Arguments:

        float_format - floating point format string"""
        return self._float_format

    def _set_float_format(self, val):
        #        self._validate_option("float_format", val)
        for field in self._field_names:
            self._float_format[field] = val

    float_format = property(_get_float_format, _set_float_format)

    def _get_padding_width(self):
        """The number of empty spaces between a column's edge and its content

        Arguments:

        padding_width - number of spaces, must be a positive integer"""
        return self._padding_width

    def _set_padding_width(self, val):
        self._validate_option('padding_width', val)
        self._padding_width = val

    padding_width = property(_get_padding_width, _set_padding_width)

    def _get_left_padding_width(self):
        """The number of empty spaces between a column's left edge and its content

        Arguments:

        left_padding - number of spaces, must be a positive integer"""
        return self._left_padding_width

    def _set_left_padding_width(self, val):
        self._validate_option('left_padding_width', val)
        self._left_padding_width = val

    left_padding_width = property(_get_left_padding_width, _set_left_padding_width)

    def _get_right_padding_width(self):
        """The number of empty spaces between a column's right edge and its content

        Arguments:

        right_padding - number of spaces, must be a positive integer"""
        return self._right_padding_width

    def _set_right_padding_width(self, val):
        self._validate_option('right_padding_width', val)
        self._right_padding_width = val

    right_padding_width = property(_get_right_padding_width, _set_right_padding_width)

    def _get_vertical_char(self):
        """The charcter used when printing table borders to draw vertical lines

        Arguments:

        vertical_char - single character string used to draw vertical lines"""
        return self._vertical_char

    def _set_vertical_char(self, val):
        val = self._unicode(val)
        self._validate_option('vertical_char', val)
        self._vertical_char = val

    vertical_char = property(_get_vertical_char, _set_vertical_char)

    def _get_horizontal_char(self):
        """The charcter used when printing table borders to draw horizontal lines

        Arguments:

        horizontal_char - single character string used to draw horizontal lines"""
        return self._horizontal_char

    def _set_horizontal_char(self, val):
        val = self._unicode(val)
        self._validate_option('horizontal_char', val)
        self._horizontal_char = val

    horizontal_char = property(_get_horizontal_char, _set_horizontal_char)

    def _get_junction_char(self):
        """The charcter used when printing table borders to draw line junctions

        Arguments:

        junction_char - single character string used to draw line junctions"""
        return self._junction_char

    def _set_junction_char(self, val):
        val = self._unicode(val)
        self._validate_option('vertical_char', val)
        self._junction_char = val

    junction_char = property(_get_junction_char, _set_junction_char)

    def _get_format(self):
        """Controls whether or not HTML tables are formatted to match styling options

        Arguments:

        format - True or False"""
        return self._format

    def _set_format(self, val):
        self._validate_option('format', val)
        self._format = val

    format = property(_get_format, _set_format)

    def _get_print_empty(self):
        """Controls whether or not empty tables produce a header and frame or just an empty string

        Arguments:

        print_empty - True or False"""
        return self._print_empty

    def _set_print_empty(self, val):
        self._validate_option('print_empty', val)
        self._print_empty = val

    print_empty = property(_get_print_empty, _set_print_empty)

    def _get_attributes(self):
        """A dictionary of HTML attribute name/value pairs to be included in the <table> tag when printing HTML

        Arguments:

        attributes - dictionary of attributes"""
        return self._attributes

    def _set_attributes(self, val):
        self._validate_option('attributes', val)
        self._attributes = val

    attributes = property(_get_attributes, _set_attributes)

    ##############################
    # OPTION MIXER               #
    ##############################

    def _get_options(self, kwargs):

        options = {}
        for option in self._options:
            if option in kwargs:
                self._validate_option(option, kwargs[option])
                options[option] = kwargs[option]
            else:
                options[option] = getattr(self, '_' + option)
        return options

    ##############################
    # PRESET STYLE LOGIC         #
    ##############################

    def set_style(self, style):

        if style == DEFAULT:
            self._set_default_style()
        elif style == MSWORD_FRIENDLY:
            self._set_msword_style()
        elif style == PLAIN_COLUMNS:
            self._set_columns_style()
        elif style == RANDOM:
            self._set_random_style()
        else:
            raise Exception('Invalid pre-set style!')

    def _set_default_style(self):

        self.header = True
        self.border = True
        self._hrules = FRAME
        self._vrules = ALL
        self.padding_width = 1
        self.left_padding_width = 1
        self.right_padding_width = 1
        self.vertical_char = '|'
        self.horizontal_char = '-'
        self.junction_char = '+'

    def _set_msword_style(self):

        self.header = True
        self.border = True
        self._hrules = NONE
        self.padding_width = 1
        self.left_padding_width = 1
        self.right_padding_width = 1
        self.vertical_char = '|'

    def _set_columns_style(self):

        self.header = True
        self.border = False
        self.padding_width = 1
        self.left_padding_width = 0
        self.right_padding_width = 8

    def _set_random_style(self):

        # Just for fun!
        self.header = random.choice((True, False))
        self.border = random.choice((True, False))
        self._hrules = random.choice((ALL, FRAME, HEADER, NONE))
        self._vrules = random.choice((ALL, FRAME, NONE))
        self.left_padding_width = random.randint(0, 5)
        self.right_padding_width = random.randint(0, 5)
        self.vertical_char = random.choice('~!@#$%^&*()_+|-={}[];\':",./;<>?')
        self.horizontal_char = random.choice('~!@#$%^&*()_+|-={}[];\':",./;<>?')
        self.junction_char = random.choice('~!@#$%^&*()_+|-={}[];\':",./;<>?')

    ##############################
    # DATA INPUT METHODS         #
    ##############################

    def add_row(self, row):

        """Add a row to the table

        Arguments:

        row - row of data, should be a list with as many elements as the table
        has fields"""

        if self._field_names and len(row) != len(self._field_names):
            raise Exception(
                'Row has incorrect number of values, (actual) %d!=%d (expected)'
                % (len(row), len(self._field_names))
            )
        if not self._field_names:
            self.field_names = [('Field %d' % (n + 1)) for n in range(0, len(row))]
        self._rows.append(list(row))

    def del_row(self, row_index):

        """Delete a row to the table

        Arguments:

        row_index - The index of the row you want to delete.  Indexing starts at 0."""

        if row_index > len(self._rows) - 1:
            raise Exception(
                'Cant delete row at index %d, table only has %d rows!'
                % (row_index, len(self._rows))
            )
        del self._rows[row_index]

    def add_column(self, fieldname, column, align='c', valign='t'):

        """Add a column to the table.

        Arguments:

        fieldname - name of the field to contain the new column of data
        column - column of data, should be a list with as many elements as the
        table has rows
        align - desired alignment for this column - "l" for left, "c" for centre and "r" for right
        valign - desired vertical alignment for new columns - "t" for top, "m" for middle and "b" for bottom"""

        if len(self._rows) in (0, len(column)):
            self._validate_align(align)
            self._validate_valign(valign)
            self._field_names.append(fieldname)
            self._align[fieldname] = align
            self._valign[fieldname] = valign
            for i in range(0, len(column)):
                if len(self._rows) < i + 1:
                    self._rows.append([])
                self._rows[i].append(column[i])
        else:
            raise Exception(
                'Column length %d does not match number of rows %d!'
                % (len(column), len(self._rows))
            )

    def clear_rows(self):

        """Delete all rows from the table but keep the current field names"""

        self._rows = []

    def clear(self):

        """Delete all rows and field names from the table, maintaining nothing but styling options"""

        self._rows = []
        self._field_names = []
        self._widths = []

    ##############################
    # MISC PUBLIC METHODS        #
    ##############################

    def copy(self):
        return copy.deepcopy(self)

    ##############################
    # MISC PRIVATE METHODS       #
    ##############################

    def _format_value(self, field, value):
        if isinstance(value, int) and field in self._int_format:
            value = self._unicode(('%%%sd' % self._int_format[field]) % value)
        elif isinstance(value, float) and field in self._float_format:
            value = self._unicode(('%%%sf' % self._float_format[field]) % value)
        return self._unicode(value)

    def _compute_widths(self, rows, options):
        if options['header']:
            widths = [_get_size(field)[0] for field in self._field_names]
        else:
            widths = len(self.field_names) * [0]
        for row in rows:
            for index, value in enumerate(row):
                fieldname = self.field_names[index]
                if fieldname in self.max_width:
                    widths[index] = max(
                        widths[index],
                        min(_get_size(value)[0], self.max_width[fieldname]),
                    )
                else:
                    widths[index] = max(widths[index], _get_size(value)[0])
        self._widths = widths

    def _get_padding_widths(self, options):

        if options['left_padding_width'] is not None:
            lpad = options['left_padding_width']
        else:
            lpad = options['padding_width']
        if options['right_padding_width'] is not None:
            rpad = options['right_padding_width']
        else:
            rpad = options['padding_width']
        return lpad, rpad

    def _get_rows(self, options):
        """Return only those data rows that should be printed, based on slicing and sorting.

        Arguments:

        options - dictionary of option settings."""

        # Make a copy of only those rows in the slice range
        rows = copy.deepcopy(self._rows[options['start'] : options['end']])
        # Sort if necessary
        if options['sortby']:
            sortindex = self._field_names.index(options['sortby'])
            # Decorate
            rows = [[row[sortindex]] + row for row in rows]
            # Sort
            rows.sort(reverse=options['reversesort'], key=options['sort_key'])
            # Undecorate
            rows = [row[1:] for row in rows]
        return rows

    def _format_row(self, row, options):
        return [
            self._format_value(field, value)
            for (field, value) in zip(self._field_names, row)
        ]

    def _format_rows(self, rows, options):
        return [self._format_row(row, options) for row in rows]

    ##############################
    # PLAIN TEXT STRING METHODS  #
    ##############################

    def get_string(self, **kwargs):

        """Return string representation of table in current state.

        Arguments:

        start - index of first data row to include in output
        end - index of last data row to include in output PLUS ONE (list slice style)
        fields - names of fields (columns) to include
        header - print a header showing field names (True or False)
        border - print a border around the table (True or False)
        hrules - controls printing of horizontal rules after rows.  Allowed values: ALL, FRAME, HEADER, NONE
        vrules - controls printing of vertical rules between columns.  Allowed values: FRAME, ALL, NONE
        int_format - controls formatting of integer data
        float_format - controls formatting of floating point data
        padding_width - number of spaces on either side of column data (only used if left and right paddings are None)
        left_padding_width - number of spaces on left hand side of column data
        right_padding_width - number of spaces on right hand side of column data
        vertical_char - single character string used to draw vertical lines
        horizontal_char - single character string used to draw horizontal lines
        junction_char - single character string used to draw line junctions
        sortby - name of field to sort rows by
        sort_key - sorting key function, applied to data points before sorting
        reversesort - True or False to sort in descending or ascending order
        print empty - if True, stringify just the header for an empty table, if False return an empty string """

        options = self._get_options(kwargs)

        lines = []

        # Don't think too hard about an empty table
        # Is this the desired behaviour?  Maybe we should still print the header?
        if self.rowcount == 0 and (not options['print_empty'] or not options['border']):
            return ''

        # Get the rows we need to print, taking into account slicing, sorting, etc.
        rows = self._get_rows(options)

        # Turn all data in all rows into Unicode, formatted as desired
        formatted_rows = self._format_rows(rows, options)

        # Compute column widths
        self._compute_widths(formatted_rows, options)

        # Add header or top of border
        self._hrule = self._stringify_hrule(options)
        if options['header']:
            lines.append(self._stringify_header(options))
        elif options['border'] and options['hrules'] in (ALL, FRAME):
            lines.append(self._hrule)

        # Add rows
        for row in formatted_rows:
            lines.append(self._stringify_row(row, options))

        # Add bottom of border
        if options['border'] and options['hrules'] == FRAME:
            lines.append(self._hrule)

        return self._unicode('\n').join(lines)

    def _stringify_hrule(self, options):

        if not options['border']:
            return ''
        lpad, rpad = self._get_padding_widths(options)
        if options['vrules'] in (ALL, FRAME):
            bits = [options['junction_char']]
        else:
            bits = [options['horizontal_char']]
        # For tables with no data or fieldnames
        if not self._field_names:
            bits.append(options['junction_char'])
            return ''.join(bits)
        for field, width in zip(self._field_names, self._widths):
            if options['fields'] and field not in options['fields']:
                continue
            bits.append((width + lpad + rpad) * options['horizontal_char'])
            if options['vrules'] == ALL:
                bits.append(options['junction_char'])
            else:
                bits.append(options['horizontal_char'])
        if options['vrules'] == FRAME:
            bits.pop()
            bits.append(options['junction_char'])
        return ''.join(bits)

    def _stringify_header(self, options):

        bits = []
        lpad, rpad = self._get_padding_widths(options)
        if options['border']:
            if options['hrules'] in (ALL, FRAME):
                bits.append(self._hrule)
                bits.append('\n')
            if options['vrules'] in (ALL, FRAME):
                bits.append(options['vertical_char'])
            else:
                bits.append(' ')
        # For tables with no data or field names
        if not self._field_names:
            if options['vrules'] in (ALL, FRAME):
                bits.append(options['vertical_char'])
            else:
                bits.append(' ')
        for field, width, in zip(self._field_names, self._widths):
            if options['fields'] and field not in options['fields']:
                continue
            if self._header_style == 'cap':
                fieldname = field.capitalize()
            elif self._header_style == 'title':
                fieldname = field.title()
            elif self._header_style == 'upper':
                fieldname = field.upper()
            elif self._header_style == 'lower':
                fieldname = field.lower()
            else:
                fieldname = field
            bits.append(
                ' ' * lpad
                + self._justify(fieldname, width, self._align[field])
                + ' ' * rpad
            )
            if options['border']:
                if options['vrules'] == ALL:
                    bits.append(options['vertical_char'])
                else:
                    bits.append(' ')
        # If vrules is FRAME, then we just appended a space at the end
        # of the last field, when we really want a vertical character
        if options['border'] and options['vrules'] == FRAME:
            bits.pop()
            bits.append(options['vertical_char'])
        if options['border'] and options['hrules'] != NONE:
            bits.append('\n')
            bits.append(self._hrule)
        return ''.join(bits)

    def _stringify_row(self, row, options):

        for index, field, value, width, in zip(
            list(range(0, len(row))), self._field_names, row, self._widths
        ):
            # Enforce max widths
            lines = value.split('\n')
            new_lines = []
            for line in lines:
                if _str_block_width(line) > width:
                    line = textwrap.fill(line, width)
                new_lines.append(line)
            lines = new_lines
            value = '\n'.join(lines)
            row[index] = value

        row_height = 0
        for c in row:
            h = _get_size(c)[1]
            if h > row_height:
                row_height = h

        bits = []
        lpad, rpad = self._get_padding_widths(options)
        for y in range(0, row_height):
            bits.append([])
            if options['border']:
                if options['vrules'] in (ALL, FRAME):
                    bits[y].append(self.vertical_char)
                else:
                    bits[y].append(' ')

        for field, value, width, in zip(self._field_names, row, self._widths):

            valign = self._valign[field]
            lines = value.split('\n')
            dHeight = row_height - len(lines)
            if dHeight:
                if valign == 'm':
                    lines = (
                        [''] * int(dHeight / 2)
                        + lines
                        + [''] * (dHeight - int(dHeight / 2))
                    )
                elif valign == 'b':
                    lines = [''] * dHeight + lines
                else:
                    lines = lines + [''] * dHeight

            y = 0
            for l in lines:
                if options['fields'] and field not in options['fields']:
                    continue

                bits[y].append(
                    ' ' * lpad
                    + self._justify(l, width, self._align[field])
                    + ' ' * rpad
                )
                if options['border']:
                    if options['vrules'] == ALL:
                        bits[y].append(self.vertical_char)
                    else:
                        bits[y].append(' ')
                y += 1

        # If vrules is FRAME, then we just appended a space at the end
        # of the last field, when we really want a vertical character
        for y in range(0, row_height):
            if options['border'] and options['vrules'] == FRAME:
                bits[y].pop()
                bits[y].append(options['vertical_char'])

        if options['border'] and options['hrules'] == ALL:
            bits[row_height - 1].append('\n')
            bits[row_height - 1].append(self._hrule)

        for y in range(0, row_height):
            bits[y] = ''.join(bits[y])

        return '\n'.join(bits)

    ##############################
    # HTML STRING METHODS        #
    ##############################

    def get_html_string(self, **kwargs):

        """Return string representation of HTML formatted version of table in current state.

        Arguments:

        start - index of first data row to include in output
        end - index of last data row to include in output PLUS ONE (list slice style)
        fields - names of fields (columns) to include
        header - print a header showing field names (True or False)
        border - print a border around the table (True or False)
        hrules - controls printing of horizontal rules after rows.  Allowed values: ALL, FRAME, HEADER, NONE
        vrules - controls printing of vertical rules between columns.  Allowed values: FRAME, ALL, NONE
        int_format - controls formatting of integer data
        float_format - controls formatting of floating point data
        padding_width - number of spaces on either side of column data (only used if left and right paddings are None)
        left_padding_width - number of spaces on left hand side of column data
        right_padding_width - number of spaces on right hand side of column data
        sortby - name of field to sort rows by
        sort_key - sorting key function, applied to data points before sorting
        attributes - dictionary of name/value pairs to include as HTML attributes in the <table> tag
        xhtml - print <br/> tags if True, <br> tags if false"""

        options = self._get_options(kwargs)

        if options['format']:
            string = self._get_formatted_html_string(options)
        else:
            string = self._get_simple_html_string(options)

        return string

    def _get_simple_html_string(self, options):

        lines = []
        if options['xhtml']:
            linebreak = '<br/>'
        else:
            linebreak = '<br>'

        open_tag = []
        open_tag.append('<table')
        if options['attributes']:
            for attr_name in options['attributes']:
                open_tag.append(
                    ' %s="%s"' % (attr_name, options['attributes'][attr_name])
                )
        open_tag.append('>')
        lines.append(''.join(open_tag))

        # Headers
        if options['header']:
            lines.append('    <tr>')
            for field in self._field_names:
                if options['fields'] and field not in options['fields']:
                    continue
                lines.append(
                    '        <th>%s</th>' % escape(field).replace('\n', linebreak)
                )
            lines.append('    </tr>')

        # Data
        rows = self._get_rows(options)
        formatted_rows = self._format_rows(rows, options)
        for row in formatted_rows:
            lines.append('    <tr>')
            for field, datum in zip(self._field_names, row):
                if options['fields'] and field not in options['fields']:
                    continue
                lines.append(
                    '        <td>%s</td>' % escape(datum).replace('\n', linebreak)
                )
            lines.append('    </tr>')

        lines.append('</table>')

        return self._unicode('\n').join(lines)

    def _get_formatted_html_string(self, options):

        lines = []
        lpad, rpad = self._get_padding_widths(options)
        if options['xhtml']:
            linebreak = '<br/>'
        else:
            linebreak = '<br>'

        open_tag = []
        open_tag.append('<table')
        if options['border']:
            if options['hrules'] == ALL and options['vrules'] == ALL:
                open_tag.append(' frame="box" rules="all"')
            elif options['hrules'] == FRAME and options['vrules'] == FRAME:
                open_tag.append(' frame="box"')
            elif options['hrules'] == FRAME and options['vrules'] == ALL:
                open_tag.append(' frame="box" rules="cols"')
            elif options['hrules'] == FRAME:
                open_tag.append(' frame="hsides"')
            elif options['hrules'] == ALL:
                open_tag.append(' frame="hsides" rules="rows"')
            elif options['vrules'] == FRAME:
                open_tag.append(' frame="vsides"')
            elif options['vrules'] == ALL:
                open_tag.append(' frame="vsides" rules="cols"')
        if options['attributes']:
            for attr_name in options['attributes']:
                open_tag.append(
                    ' %s="%s"' % (attr_name, options['attributes'][attr_name])
                )
        open_tag.append('>')
        lines.append(''.join(open_tag))

        # Headers
        if options['header']:
            lines.append('    <tr>')
            for field in self._field_names:
                if options['fields'] and field not in options['fields']:
                    continue
                lines.append(
                    '        <th style="padding-left: %dem; padding-right: %dem; text-align: center">%s</th>'
                    % (lpad, rpad, escape(field).replace('\n', linebreak))
                )
            lines.append('    </tr>')

        # Data
        rows = self._get_rows(options)
        formatted_rows = self._format_rows(rows, options)
        aligns = []
        valigns = []
        for field in self._field_names:
            aligns.append(
                {'l': 'left', 'r': 'right', 'c': 'center'}[self._align[field]]
            )
            valigns.append(
                {'t': 'top', 'm': 'middle', 'b': 'bottom'}[self._valign[field]]
            )
        for row in formatted_rows:
            lines.append('    <tr>')
            for field, datum, align, valign in zip(
                self._field_names, row, aligns, valigns
            ):
                if options['fields'] and field not in options['fields']:
                    continue
                lines.append(
                    '        <td style="padding-left: %dem; padding-right: %dem; text-align: %s; vertical-align: %s">%s</td>'
                    % (
                        lpad,
                        rpad,
                        align,
                        valign,
                        escape(datum).replace('\n', linebreak),
                    )
                )
            lines.append('    </tr>')
        lines.append('</table>')

        return self._unicode('\n').join(lines)


##############################
# UNICODE WIDTH FUNCTIONS    #
##############################


def _char_block_width(char):
    # Basic Latin, which is probably the most common case
    # if char in xrange(0x0021, 0x007e):
    # if char >= 0x0021 and char <= 0x007e:
    if 0x0021 <= char <= 0x007E:
        return 1
    # Chinese, Japanese, Korean (common)
    if 0x4E00 <= char <= 0x9FFF:
        return 2
    # Hangul
    if 0xAC00 <= char <= 0xD7AF:
        return 2
    # Combining?
    if unicodedata.combining(uni_chr(char)):
        return 0
    # Hiragana and Katakana
    if 0x3040 <= char <= 0x309F or 0x30A0 <= char <= 0x30FF:
        return 2
    # Full-width Latin characters
    if 0xFF01 <= char <= 0xFF60:
        return 2
    # CJK punctuation
    if 0x3000 <= char <= 0x303E:
        return 2
    # Backspace and delete
    if char in (0x0008, 0x007F):
        return -1
    # Other control characters
    elif char in (0x0000, 0x001F):
        return 0
    # Take a guess
    return 1


def _str_block_width(val):

    return sum(map(_char_block_width, list(map(ord, _re.sub('', val)))))


##############################
# TABLE FACTORIES            #
##############################


def from_csv(fp, field_names=None, **kwargs):

    dialect = csv.Sniffer().sniff(fp.read(1024))
    fp.seek(0)
    reader = csv.reader(fp, dialect)

    table = PrettyTable(**kwargs)
    if field_names:
        table.field_names = field_names
    else:
        if py3k:
            table.field_names = [x.strip() for x in next(reader)]
        else:
            table.field_names = [x.strip() for x in next(reader)]

    for row in reader:
        table.add_row([x.strip() for x in row])

    return table


def from_db_cursor(cursor, **kwargs):

    if cursor.description:
        table = PrettyTable(**kwargs)
        table.field_names = [col[0] for col in cursor.description]
        for row in cursor.fetchall():
            table.add_row(row)
        return table


class TableHandler(HTMLParser):
    def __init__(self, **kwargs):
        HTMLParser.__init__(self)
        self.kwargs = kwargs
        self.tables = []
        self.last_row = []
        self.rows = []
        self.max_row_width = 0
        self.active = None
        self.last_content = ''
        self.is_last_row_header = False

    def handle_starttag(self, tag, attrs):
        self.active = tag
        if tag == 'th':
            self.is_last_row_header = True

    def handle_endtag(self, tag):
        if tag in ['th', 'td']:
            stripped_content = self.last_content.strip()
            self.last_row.append(stripped_content)
        if tag == 'tr':
            self.rows.append((self.last_row, self.is_last_row_header))
            self.max_row_width = max(self.max_row_width, len(self.last_row))
            self.last_row = []
            self.is_last_row_header = False
        if tag == 'table':
            table = self.generate_table(self.rows)
            self.tables.append(table)
            self.rows = []
        self.last_content = ' '
        self.active = None

    def handle_data(self, data):
        self.last_content += data

    def generate_table(self, rows):
        """
        Generates from a list of rows a PrettyTable object.
        """
        table = PrettyTable(**self.kwargs)
        for row in self.rows:
            if len(row[0]) < self.max_row_width:
                appends = self.max_row_width - len(row[0])
                for i in range(1, appends):
                    row[0].append('-')

            if row[1] == True:
                self.make_fields_unique(row[0])
                table.field_names = row[0]
            else:
                table.add_row(row[0])
        return table

    def make_fields_unique(self, fields):
        """
        iterates over the row and make each field unique
        """
        for i in range(0, len(fields)):
            for j in range(i + 1, len(fields)):
                if fields[i] == fields[j]:
                    fields[j] += "'"


def from_html(html_code, **kwargs):
    """
    Generates a list of PrettyTables from a string of HTML code. Each <table> in
    the HTML becomes one PrettyTable object.
    """

    parser = TableHandler(**kwargs)
    parser.feed(html_code)
    return parser.tables


def from_html_one(html_code, **kwargs):
    """
    Generates a PrettyTables from a string of HTML code which contains only a
    single <table>
    """

    tables = from_html(html_code, **kwargs)
    try:
        assert len(tables) == 1
    except AssertionError:
        raise Exception(
            'More than one <table> in provided HTML code!  Use from_html instead.'
        )
    return tables[0]


##############################
# MAIN (TEST FUNCTION)       #
##############################


def main():

    x = PrettyTable(['City name', 'Area', 'Population', 'Annual Rainfall'])
    x.sortby = 'Population'
    x.reversesort = True
    x.int_format['Area'] = '04d'
    x.float_format = '6.1f'
    x.align['City name'] = 'l'  # Left align city names
    x.add_row(['Hoehenkirchen\nSiegertsbrunn', 1295, 1158259, 600.5])
    x.add_row(['Adelaide', 1295, 1158259, 600.5])
    x.add_row(['Brisbane', 5905, 1857594, 1146.4])
    x.add_row(['Darwin', 112, 120900, 1714.7])
    x.add_row(['Hobart', 1357, 205556, 619.5])
    x.add_row(['Sydney', 2058, 4336374, 1214.8])
    x.add_row(['Melbourne', 1566, 3806092, 646.9])
    x.add_row(['Perth', 5386, 1554769, 869.4])
    print(x)


if __name__ == '__main__':
    main()



"""
Parser for a Type / Class Tree

After a call to build_type_hirarchy, the types are put into a hirarchy, then
the class parsing starts.

The main features this thing provides is to detect type derived superclasses and
rearanges of them as new classes within the defining one.

        class Foo(Bar, Baz) ->
        class Foo(Bar)
                class Baz(Baz)

where in this not yet built when doing this,
we are in its metacls __new__ yet.

later Foo.Baz can be getting special settings without influencing the other
Baz classes.

This allows for very small definition specs.

"""

# -------------------------------------------------------------- Build the Tree
hirarchy = []
hirarchy_built = False
# non interesting bases, up to the first type:
mixins = {}
classes = {}
cls_counter = 0


def out(*msg):
    try:
        console.log(' '.join([str(m) for m in msg]))
    except:
        print(*msg)


class MixinTypeMetaClass(type):
    """ outside the hirarchy but still a type.
    not in use currently, was e.g. for Label types like ServiceGroup=foo
    (later we added svc group into the hirarchy)
    """

    def __new__(cls, name, bases, attrs):
        ncls = type.__new__(cls, name, bases, attrs)
        mixins.setdefault(ncls, [])
        return ncls


class MT(object):
    __metaclass__ = MixinTypeMetaClass


def build_type_hirarchy(root):
    """ e.g. Project"""
    out(hirarchy, 'fooo')
    if root not in hirarchy:
        raise Exception(root, 'not in hirarchy')
    while 1:
        if hirarchy[0] != root:
            hirarchy.pop(0)
        else:
            break
    out('have hirarchy', [n.__name__ for n in hirarchy])
    global hirarchy_built
    hirarchy_built = True  # close hirachy, following are specific classes


def add_to(*dests, **kw):
    'convenience wrapper for add_cls - from spec'
    dests = _list(dests)
    for parent in dests:
        for k, childs in kw.items():
            for child in _list(childs):
                if not child.type == k:
                    raise Exception("Can't add: %s is no %s" % (child, k))
                add_cls(parent, child)


def add_cls(parent, child):
    """ called from the spec """
    check_allow_add(parent, child)
    out('    adding %s to %s' % (child, parent))
    new_sub = type(child.__name__, (child,), {'_parent': parent})
    setattr(parent, child.__name__, new_sub)
    return new_sub


class HirarchyErr(Exception):
    pass


def check_allow_add(cls, sub):
    if not sub._hirarchy == cls._hirarchy + 1:
        for c in cls._allowed_childs:
            if sub.type == c.type:
                return
        raise HirarchyErr("Can't add %s directly within %s" % (sub, cls))


# ----------------------------------------------------------------------------


class TypeMetaClass(type):
    def __new__(cls, name, bases, attrs):
        attrs = dict(attrs)
        b0, orig_bases = bases[0], None
        if b0 in mixins and not hirarchy_built:
            raise Exception("First Base Class can't be a mixin", b0)
        if hirarchy_built:
            if not '_parent' in attrs:
                while not b0 in hirarchy:
                    # a helper class (like class Node(Node, Role.Foo) and inner Node is
                    # hirarchy type -> act as if we'd defined all:
                    inh = classes[b0]
                    b0, bases = inh[0], inh + bases[1:]
                attrs['type'] = b0.__name__
                attrs['cls'] = name
                out('adding %s(%s)' % (name, b0.__name__))

                bases, orig_bases, mxins, attrs = set_bases_as_attrs(
                    name, b0, bases, attrs
                )

        else:
            out('registring type', name)
            attrs['type'] = name
            attrs.setdefault('_allowed_childs', [])

        global cls_counter
        cls_counter += 1
        attrs['_nr'] = cls_counter
        ncls = type.__new__(cls, name, bases, attrs)
        if hirarchy_built:
            if '_parent' in attrs:
                return ncls
        if orig_bases:
            classes[ncls] = orig_bases
            for m in mxins:
                if hasattr(m, '__code__'):
                    m(ncls)
                else:
                    out(
                        'Adding %s to %s %s'
                        % (name, mixins[m]['orig'].__name__, m.__name__)
                    )
                    mixins[m]['members'].append(ncls)
            return ncls
        if not hirarchy_built and len(hirarchy):
            hirarchy[-1:][0]._allowed_childs.append(ncls)
            ncls._hirarchy = len(hirarchy)
        hirarchy.append(ncls)
        return ncls

    def __repr__(cls):
        g = getattr
        p = g(cls, '_parent', None)
        if isinstance(p, str):
            p = 'prebound(%s)' % p
        elif p == None:
            p = 'type'
        else:
            if not g(p, '_parent', None):
                p = ''
            else:
                p = ' [%s]' % p
        return '%s.%s.%s%s' % (cls.type, g(cls, 'cls', ''), g(cls, '_nr', 0), p)


# import sys
# if sys.version_info.major == 3:
#    exec ('class T(object, metaclass=TypeMetaClass): descr=None')
# else:
class T(object, metaclass=TypeMetaClass):
    descr = None


# class T(object):
#    __metaclass__ = TypeMetaClass
#    descr = None


simple_types = (float, int, bool, str, type(None), tuple, list, dict)


def is_sub_cls(c):
    if is_simple_type(c):
        return
    if not getattr(c, '__class__', None):
        return
    return False
    out(isinstance(3, simple_types))
    out(isinstance(c, simple_types))
    return isinstance(c, simple_types) and not isinstance(c, property)

    # return getattr(c, '_parent', None)


def descr(c):
    """up the mro to find any descr or __doc__
    a manually set descr interrupts the process.
    """
    return 'decscr'
    d = ()
    alert(type(c))
    for b in c.mro()[:-2]:
        dd = getattr(b, 'descr', None)
        if dd and not dd in d:
            d += (dd,)
            break
        else:
            dd = b.__doc__
            if dd and not dd in d:
                d += (dd,)
    return '.'.join(d)


def build_tree(ncls):
    """ configure all prebound classes, add tree attrs to the others (_parent) """
    global hirarchy_built

    def pre(c, r, hir, cfg):
        c.descr = descr(c)
        childs = unordered_childs(c, is_sub_cls)
        for cc in childs:
            cc._parent = c
            for k in '_nr', '_hirarchy':
                cc._nr = getattr(cc, k, 0)
            cc.type = getattr(cc, 'type', '%s.%s' % (c.mro()[1].__name__, cc.__name__))
            cc.cls = getattr(cc, 'cls', cc.__name__)

    walk_tree(ncls, pre, None, is_sub_cls)
    hirarchy_built = 2


oltens = []


def set_bases_as_attrs(name, b0, orig_bases, attrs):
    """ type hirarchy is built at this point
    Our job now is to pop those bases (and add them as subclasses which
    registered already).
    If registered as mixin then we create new mixins or resolve - see below
    """

    def add_local(b0, name, sub, parent_attrs, lms=None):
        """
        """
        try:
            check_allow_add(b0, sub)
        except HirarchyErr as ex:
            if not lms:
                raise Exception('Hirarchy error %s %s' % (b0, sub))

            for i in range(len(lms), 0, -1):
                l = lms[i - 1]
                try:
                    t = add_cls(l, sub)
                    return t
                except HirarchyErr:
                    lms.pop(i - 1)
            raise Exception('Cannot add local %s %s' % (b0, sub))
        bn = sub.__name__
        out('    adding %s to %s' % (sub, name))
        parent_attrs[bn] = t = type(bn, (sub,), {'_parent': name})
        parent_attrs.setdefault('_from_mro', []).append(t)
        return t

    mxins = []

    def new_mixin(ncls, b0, orig_mixin):
        if ncls in mixins:
            raise Exception('mixin %s already defined' % ncls)
        mixins[ncls] = {'members_cls': b0, 'members': [], 'orig': orig_mixin}

    mx = mixins.get(b0)
    if mx:
        # class NBI(AXESS) where AXESS is a group
        # -> we set group='AXESS' and b0 to Service
        mxins.append(b0)  # this will add this new cls to b0's members,
        # we don't have it yet
        attrs[mx['orig'].__name__] = b0.__name__
        b0 = mx['members_cls']
        attrs['type'] = b0.type

    new_bases = [b0]
    lms = []  # last added mro based sub
    for b in orig_bases[1:]:
        if b in mixins:
            # a mixin type in the bases: class AXESS(Service, Group)
            # -> we'll create a NEW mixin, but with a members_type,
            # e.g. Service
            # if such a members_cls is alredy set, then we resolve
            # e.g. class Foo(Role, AXESS), AXESS is such a mx-> resolve
            mx = mixins[b]
            if 'members_cls' in mx:
                out('    resolving all members of %s' % b.__name__)
                for m in mx['members']:
                    lms = [add_local(b0, name, sub=m, parent_attrs=attrs)]
            else:
                # create new
                mxins.append(lambda ncls: new_mixin(ncls, b0, b))
        elif b in hirarchy:
            import pdb

            pdb.set_trace()
        elif b in classes:
            lms.append(add_local(b0, name, sub=b, parent_attrs=attrs, lms=lms))
        else:
            # normal superclass:
            new_bases.append(b)
    return tuple(new_bases), orig_bases, mxins, attrs


def allow(*what, **on):
    import pdb

    pdb.set_trace()
    # if exec_env == 'server':
    #    [k._allowed_childs.append(v) for k in _list(on.pop('on')) \
    #                                for v in _list(what)]


def _list(s):
    if s and isinstance(s, tuple) and isinstance(s[0], (tuple, list)):
        s = s[0]
    return list(s) if isinstance(s, (list, tuple)) else [s]


# ------------------------------------------------------------ Explore the Tree


def props(c):
    l = [(k, getattr(c, k)) for k in dir(c) if not k.startswith('_')]
    return [(k, v) for k, v in l if not is_function(v)]


def is_function(v):
    return typeof(v) == 'function'


def is_simple_type(v):
    if v is None:
        return True


__pragma__('ifdef', 'onserver')


def is_function(v):
    return hasattr(v, '__code__')


def is_simple_type(v):
    return isinstance(v, simple_types)


__pragma__('endif')


def walk_tree(cur, pre=None, post=None, match=None, res=None, cfg=None, hir=0):
    """
    Recursing the class tree, starting from a given root node (cur).
    the result is mutated
    pre, post, match = callables
    """
    if res is None:
        res = []
    hir += 1
    if pre:
        pre(cur, res, hir, cfg)
    if not cfg or cfg.get('stop_recursion') != cur.type:
        for k in ordered_childs(cur, match):
            walk_tree(k, pre, post, match, res, cfg, hir=hir)
    if post:
        post(cur, res, hir, cfg)
    return res


def unordered_childs(parent, match):
    c = getattr(parent, '_childs', None)
    if c:
        return c
    childs = []
    for k, v in props(parent):
        if match(v):
            childs.append(v)
    parent._childs = childs
    parent._childs_ordered = False
    return childs


def ordered_childs(parent, match):
    childs = unordered_childs(parent, match)
    if getattr(parent, '_childs_ordered', 0):
        return childs
    childs.sort(key=lambda x: getattr(x, '_nr', 0))
    parent._childs_ordered = True
    return childs



from functools import partial

# fmt:off
code = lambda typ, body: '''

```%s
%s
```

''' % (typ, body)

call_seq_diag = '''


[![](.%(f)s.svg)](%(l)scall_sequence.html?src=%(f)s)


'''
# fmt:on


class Viz:
    activate = lambda abbr: 'activate %s' % abbr
    deactivate = lambda abbr: 'deactivate %s' % abbr
    req = lambda frm, to, f=lambda x: x, what='': '%s ->> %s: %s' % (frm, to, f(what))
    rsp = lambda frm, to, f=lambda x: x, what='': '%s -->> %s: %s' % (frm, to, f(what))
    note_over = lambda n, what: 'note over %s: %s' % (n, what)


class Plant(Viz):
    participant = lambda abbr, title: 'participant "%s" as %s' % (title, abbr)
    ext = 'plantuml'
    wrap = lambda r: '@startuml\n%s\n@enduml' % '\n'.join(r)
    # include = lambda fn: '\n![call sequence](.%s.svg)\n' % fn.rsplit('.', 1)[0]
    include = lambda fn, loc: call_seq_diag % {'f': fn.rsplit('.', 1)[0], 'l': loc}
    ls = '\\n'


class Mermaid(Viz):
    participant = lambda abbr, title: 'participant %s as %s' % (abbr, title)
    what = lambda w='': '""' if not w else '%s' % w
    req = partial(Viz.req, f=what)
    rsp = partial(Viz.rsp, f=what)
    ext = 'mermaid'
    wrap = lambda r: code('mermaid', 'sequenceDiagram\n%s' % '\n'.join(r))
    ls = '<br>'



try:
    pass

    have_colorama = True
except:
    have_colorama = False
import sys

from structlog.dev import ConsoleRenderer
from structlog.stdlib import ProcessorFormatter

from .renderers import ThemeableConsoleRenderer

have_stdout = sys.stdout.isatty()


class ConsoleFormatter(ProcessorFormatter):
    """
    Line logger, intended for ttys.

    This class is a helper to get json dict configurable

    Example:
     "formatters": {
        "axrx_console": {
            "()": "ax.utils.logging.structlog.ConsoleFormatter",
            "colors": true
        },

    """

    def __init__(
        self, colors='auto', theme=None, val_formatters=None, fmt_vals=None, **kw
    ):
        """we use the built in colorama based colorizer only when colors=True
        """
        if colors == 'auto':
            colors = True if have_stdout else False

        if (
            not theme
            and not fmt_vals
            and (not colors or (colors is True and have_colorama))
        ):
            # the sl default:
            r = ConsoleRenderer(colors=colors, **kw)
        else:
            # ours. no colorama dependency, themes, coloring features
            r = ThemeableConsoleRenderer(
                colors=colors,
                theme=theme,
                val_formatters=val_formatters,
                fmt_vals=fmt_vals,
                **kw
            )

        ProcessorFormatter.__init__(
            self,
            processor=r
            # , foreign_pre_chain = pre_chain
            ,
            keep_exc_info=True,
            keep_stack_info=False,  # py3 only
        )



#
# Copyright (C) 2010-2012 Vinay Sajip. All rights reserved.
# Licensed under the new BSD license.
# AX: Deleted all windows functions
#
import logging

from theming.formatters import BPFormat


class ColorizingStreamHandler(logging.StreamHandler):
    # color names to indices
    color_map = {
        'black': 0,
        'red': 1,
        'green': 2,
        'yellow': 3,
        'blue': 4,
        'magenta': 5,
        'cyan': 6,
        'white': 7,
    }

    # levels to (background, foreground, bold/intense)
    level_map = {
        logging.DEBUG: (None, 'blue', False),
        logging.INFO: (None, 'black', False),
        logging.WARNING: (None, 'yellow', False),
        logging.ERROR: (None, 'red', False),
        logging.CRITICAL: ('red', 'white', True),
    }
    csi = '\x1b['
    reset = '\x1b[0m'

    @property
    def is_tty(self):
        isatty = getattr(self.stream, 'isatty', None)
        return isatty and isatty()

    def emit(self, record):
        try:
            message = self.format(record)
            stream = self.stream
            if not self.is_tty:
                stream.write(message)
            else:
                self.output_colorized(message)
            stream.write(getattr(self, 'terminator', '\n'))
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def output_colorized(self, message):
        self.stream.write(message)

    def colorize(self, message, record):
        if record.levelno in self.level_map:
            bg, fg, bold = self.level_map[record.levelno]
            params = []
            if bg in self.color_map:
                params.append(str(self.color_map[bg] + 40))
            if fg in self.color_map:
                params.append(str(self.color_map[fg] + 30))
            if bold:
                params.append('1')
            if params:
                message = ''.join(
                    (self.csi, ';'.join(params), 'm', message, self.reset)
                )
        return message

    def format(self, record):
        message = logging.StreamHandler.format(self, record)
        if self.is_tty:
            # Don't colorize any traceback
            parts = message.split('\n', 1)
            parts[0] = self.colorize(parts[0], record)
            message = '\n'.join(parts)
        return message


class ColorizingStreamHandlerForBPython(ColorizingStreamHandler):
    def colorize(self, message, record):
        if record.levelno in self.level_map:
            bg, fg, bold = self.level_map[record.levelno]
            message = BPFormat.format(message, fg, bold)
        message = BPFormat.fmt(message)
        return message


def main():
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(ColorizingStreamHandler())
    logging.debug('DEBUG')
    logging.info('INFO')
    logging.warning('WARNING')
    logging.error('ERROR')
    logging.critical('CRITICAL')


if __name__ == '__main__':
    main()



from devapp.tools import offset_port, exists, project
import os
from devapp.app import app


class tool:
    conda_env = 'lctools'


def redis_server(**kw):
    return 'redis-server --port %s' % offset_port(6379)


def verify_tools(path, rsc, **kw):
    for p in rsc.provides:
        d = path + '/' + p
        if not exists(d):
            app.error('Not found', cmd=d)
            return False
    return True


def lc_tools(*a, **kw):
    cmd = kw['cmd']
    if cmd == 'tmux':
        d = project.root() + '/tmp/tmux'
        os.makedirs(d, exist_ok=True)
        return {'cmd_pre': 'export TMUX_TMPDIR="%s"; ' % d, 'cmd': 'tmux'}
    return cmd


class rsc:
    """For services we change dir to project."""

    class redis_server:
        cmd = 'redis-server'
        run = redis_server
        pkg = 'redis'
        port = 6379
        systemd = 'redis-server'

    class lc_tools:
        # httpd for rotatelogs
        provides = ['git', 'fzf', 'jq', 'rg', 'fd', 'http', 'htop', 'tmux']
        conda_chan = 'conda-forge'
        conda_pkg = (
            ' '.join(provides)
            .replace(' rg ', ' ripgrep ')
            .replace(' fd ', ' fd-find ')
            .replace(' http ', ' httpie ')
        )
        run = lc_tools
        verify_present = verify_tools

    class lc_tools_kf:
        # httpd for rotatelogs
        provides = ['rotatelogs']
        conda_chan = 'kalefranz'
        conda_pkg = 'httpd'
        verify_present = verify_tools



#!/usr/bin/env python
"""
This is the most concise form to build a nested class tree, where the local
classes are independent clones (can be configured w/o affecting others).

CAUTION:
We use a few tricks at class generation, inspecting the mro and put any ref
to a local class out of from the mro, into the container class as attribute!


Config: Class Tree
State: Object Tree
"""

from tree_builder import *


# there might be more hirarchies, this is the canonical one:
class Project(T):
    pass


class Cluster(T):
    pass


class Datacenter(T):
    pass


class Tier(T):
    pass


class Node(T):
    pass


class Role(T):
    pass


class Container(T):
    networking = 'host'


class SvcGroup(T):
    pass


class Service(T):
    criticality = 1


# Swisscom.MBC.SB.Dev.Zuerich.ACS1.Nginx


# mixin types. not needed. yet.
# class Group   ( MT ): short = 'grou'

build_type_hirarchy(root=Project)


allow([Service], on=(Role, Container))

# -----------------------------------------------------------------------------


class Zeo(Service):
    'Object DB for AXESS'


class MySQL(Service):
    'SQL DB for AXESS'


class Redis(Service):
    'Local KV Store, Job Queue for AXESS/AXD'


class Nginx(Service):
    'Proxy, Static Content'


class NBI(Service):
    'Tibco In'


class Callback(Service):
    'Job Responses'


class Portal(Service):
    'HET Portal'


class Debug(Service):
    'Debugging / Testing'


class Phase0(Service):
    'Unauth Field'


class Phase1(Service):
    'Authenticated Field'


class TR069(Service):
    'TR069'


class GUI(Service):
    'Admin GUI'


class Configurator(Service):
    'TR069 ConnRequests Out'


class AXD_A1(Service):
    'Broadsoft Ocip / Tibco Out'


class AXD_A2(Service):
    'Genband Out'


class AXD_System(Service):
    'Local Jobs'


class Notebook(Service):
    'Interactive NBI'


class Repo(Service):
    'Repositories'


class ConsulClient(Service):
    'Local HealthChecker / Reporter'


class ConsulServer(Service):
    'Datacenter Wide Truth Provider / KV Store'


class L0Master(Service):
    'Triggers Cluster L0 Jobs, RPC GUI'


class L0Agent(Service):
    'Runs Cluster L0 Jobs'


class DCStatus(Service):
    'Final Datacenter Status Based on All Results'


class AXESS(
    SvcGroup, NBI, Callback, Portal, Debug, Phase0, Phase1, TR069, GUI, Configurator,
):
    'HTTP App Server'


class AXD(SvcGroup, AXD_A1, AXD_A2, AXD_System):
    'Outgoing Job Runner'


# common props:
DCStatus.interval = ConsulClient.interval = 5
AXESS.Debug.criticality = 3
ConsulClient.criticality = 3
Notebook.criticality = ConsulServer.criticality = DCStatus.criticality = 2
L0Agent.criticality = L0Master.criticality = Repo.criticality = 2


class ContNB(Container, AXESS, AXD, Notebook, Nginx):
    pass


class Cont:
    'NSpawn Virtual Nodes'

    class ACS(Container, Nginx):
        'DevMgmt'

    class NB_MBC(ContNB):
        'Stateless MBC'

    class NB_SME(ContNB):
        'Stateless SME'

    class NB_ST(ContNB):
        'Stateless Siptrunk'

    class DB(Container, Zeo, MySQL, Redis):
        'Database / State'


class Role:
    class L0(Role, L0Agent):
        'Operational Layer / Cluster O&M Jobs'

    class L1(Role, ConsulClient):
        'Consul Layer'

    class MBC(Role):
        'SmartBCon'

    class SME(Role):
        'SmallMediumEnterprise'

    class ST(Role):
        'SipTrunk'


# helper
class N(Node, Role.L0, Role.L1):
    'Phys. Node'


class ACS(N, Role.MBC, Cont.ACS, Role.SME, Cont.ACS):
    'MBC Field'


class M(N, Role.MBC, Cont.NB_MBC, Cont.DB):
    'MBC NB'


class S(N, Role.SME, Cont.NB_SME, Cont.DB):
    'SME NB'


class E(N, Role.ST, Cont.NB_ST, Cont.DB):
    'ST  NB'


class C(N):
    'O&M Master'


add_to(C.L0, Service=(L0Master, Repo))
add_to(M.L1, S.L1, E.L1, Service=[ConsulServer, DCStatus])


class NB(Tier, M, S, E):
    'Northbound Tier'


class SB(Tier, ACS):
    'Southbound Tier'


class L0Master(Tier, C):
    'O&M'


class Zuerich(Datacenter, L0Master, NB, SB):
    'Primary'


class Olten(Datacenter, L0Master, NB, SB):
    'HotStandby'


class Cluster(Cluster, Zuerich, Olten):
    'Cluster'


class Dev(Cluster):
    'Development'


class A(Cluster):
    'Staging'


class B(Cluster):
    'PreProduction'


class Prod(Cluster):
    'Production'


class Swisscom(Project, Dev, A, B, Prod):
    'Enterprise Voice'


postconfigure(Swisscom)

# ------------------------------------------------------------------ Properties
Swisscom.Dev.Zuerich.L0Master.C.ip = '10.67.225.109'
Swisscom.Dev.Zuerich.NB.M.ip = '10.67.225.103'
Swisscom.Dev.Zuerich.NB.S.ip = '10.67.225.105'
Swisscom.Dev.Zuerich.NB.E.ip = '10.67.225.107'
Swisscom.Dev.Zuerich.SB.ACS.ip = '195.186.149.150'
Swisscom.Dev.Olten.L0Master.C.ip = '10.67.224.108'
Swisscom.Dev.Olten.NB.M.ip = '10.67.224.102'
Swisscom.Dev.Olten.NB.S.ip = '10.67.224.104'
Swisscom.Dev.Olten.NB.E.ip = '10.67.224.106'
Swisscom.Dev.Olten.SB.ACS.ip = '195.186.149.152'


if __name__ == '__main__':
    import sys

    sys.argv.append('')
    from .render import to_html

    # print to_html(Swisscom.A.Zuerich.SB.ACS, file=sys.argv[1], into='<!--cl-->')
    # print to_html(Swisscom.Dev, file=sys.argv[1], into='<!--cl-->',
    #        props_ign=('cls', 'descr', 'name', 'type', 'criticality'),
    #        stop_recursion='Role', no_props=1)
    print(
        to_html(
            Swisscom.Dev,
            file=sys.argv[1],
            into='<!--cl-->',
            props_ign=('cls', 'descr', 'name', 'type', 'criticality'),
        )
    )



#!/usr/bin/env python
"""
Building Containers


Merging filesystem layers
- either from dpull (oci compliant)
- plain dirs

using tar -> copy over
using union2 -> mount overlay(2)
"""

import json
import os
import socket
import time

from devapp.app import FLG, app, run_app
from mdvl.tools import add_metas, struct_code_block

exists = os.path.exists
abspath = os.path.abspath

root = os.geteuid() == 0


class Flags:
    autoshort = ''

    class skip_filesystem_adaptions:
        n = 'Do not run any FS changes, e.g. setting root pw. Default is True if we are not root.'
        d = False if root else True

    class write_markdown:
        n = 'Write infos into this file md formatted'

    class dirs:
        n = ('Exisiting layers or plain dirs - merged in comma sep. order given',)
        t = list
        d = ['./images']

    class target_dir:
        n = 'target filesystem directory'

    class make_tar:
        n = 'Pack together target dir into this file'

    class method:
        n = 'merge method. union requires overlay2. Supported: tar, union'
        d = 'tar'

    class rootpw:
        n = 'set root password (empty to not change)'
        d = 'devapp' if root else ''

    class allow_pts_logins:
        n = 'Allow machinectl logins for so many pts (0: skip this)'
        d = 8 if root else 0

    class allow_console_login:
        n = 'Allow console login when booted (via nspawn)'
        d = True if root else False

    class add_meta_data:
        n = 'Store meta data into target'
        d = True

    class tar_excludes:
        n = 'Add tar excludes as given'
        t = list
        d = [] if root else ['dev/*']


F = FLG


def die(msg, **kw):
    app.die(msg, **kw)


def run_cmd(cmd, msg='Running', no_fail=False):
    app.log.info(msg, cmd=cmd)
    res = os.system(cmd)
    if res:
        if no_fail:
            return res
        die('Failed - exitting')


def add_layer(ds, lay):
    app.log.info('Adding', layer=lay)
    dt = F.target_dir
    excl = FLG.tar_excludes or ''
    if excl:
        excl = ' '.join([' --exclude=%s ' % i for i in excl])

    if F.method == 'tar':
        run_cmd('tar -C  "%s/" %s --force-local -xf "%s"' % (dt, excl, ds + '/' + lay))
        return
    raise NotImplemented


def run():
    if not root:
        FLG.skip_filesystem_adaptions = True
    if not FLG.skip_filesystem_adaptions:
        run_cmd('chroot / sleep 0', 'Checking: Have chroot')
    if F.method == 'tar':
        run_cmd('tar --version | head -n1', 'Checking: Have tar')
    ds = F.dirs
    if not ds:
        die('No source dir')
    ds = F.dirs = [s[:-1] if s.endswith('/') else s for s in ds]

    if F.write_markdown and not F.add_meta_data:
        die('To write markdown we need to collect metadata')

    for d in ds:
        if not exists(d):
            die('Not found', dir=d)
        if d == F.target_dir:
            die('Source dir is target dir', dir=d)

    if F.make_tar and exists(F.make_tar):
        die('Exists', tar=F.make_tar)
    _ = F.write_markdown
    if _:
        F.write_markdown = _ = abspath(_)
        if not exists(_):
            app.log.info('Creating', file=_)
            run_cmd('mkdir -p "%s" && touch "%s"' % (_.rsplit('/', 1)[0], _))

    dt = F.target_dir
    dt = abspath(dt) if dt else ''
    dtar = abspath(F.make_tar) if F.make_tar else ''
    rm_work_dir = False
    if not dt and dtar:
        dt = abspath(ds[0]) + '.work'
        rm_work_dir = True
    else:
        if len(dt.split('/')) < 3 or dt == '/':
            die('Refusing (too short)', target_dir=dt)
    F.target_dir = dt

    for d in ds:
        do_dir(d)
    if not FLG.skip_filesystem_adaptions:
        adapt_fs()

    if F.add_meta_data:
        add_meta_data()

    if F.write_markdown:
        write_markdown()

    if dtar:
        do_tar(dtar)
    msg = 'Successfully created filesystem'
    if not FLG.skip_filesystem_adaptions:
        app.log.info(
            msg + '. Chroot uname output',
            out=os.popen('chroot "%s" uname -a' % dt).read(),
        )
    else:
        app.log.info(msg, dir=dt)

    if rm_work_dir:
        run_cmd('rm -rf "%s"' % dt, 'Removing work dir')


def do_tar(dtar):
    app.log.info('------------ BUILDING TAR ---------------')
    run_cmd('mkdir -p "%s"' % dtar.rsplit('/', 1)[0])
    run_cmd('( cd "%s/" && tar -cf "%s" . )' % (F.target_dir, dtar))


def adapt_fs():
    app.log.info('------------ ADAPT FILESYSTEM -----------')
    rp, dt = F.rootpw, F.target_dir
    if rp:
        r = run_cmd(
            "echo 'root:%s' | chroot '%s' chpasswd" % (rp, dt),
            'Changing root password',
            no_fail=True,
        )
        if r:
            app.log.error('Could not set password. chpasswd utility missing')
    nrc = F.allow_pts_logins
    with open(dt + '/etc/securetty') as fd:
        s = fd.read().splitlines()
    lines = [l for l in s if not 'pts' in l and not 'console' in l]
    for i in range(0, nrc):
        lines.insert(0, 'pts/%s' % i)
        app.log.info('Adding pts', nr=i)
    if F.allow_console_login:
        lines.insert(0, 'console')
        app.log.info('Adding console login')
    with open(dt + '/etc/securetty', 'w') as fd:
        fd.write('\n'.join(lines))


def find_files(d, match):
    return (
        os.popen('cd "%s" && find . -print |grep "%s"' % (d, match)).read().splitlines()
    )


def read(fn):
    with open(fn) as fd:
        s = fd.read()
        try:
            return json.loads(s)
        except:
            return s


def document(d, md):
    app.log.info('Documenting', d=d)
    for mani in find_files(d, 'manifestJson'):
        fmani = d + '/' + mani
        mani = read(fmani)
        for l in mani:
            md.append('RepoTags: %s' % ', '.join(l['RepoTags']))
            cfg = read(fmani.replace('manifestJson', l['Config']))
            md.append(struct_code_block(cfg, title='Config'))

            md.append('### Layers:')
            for L in l['Layers']:
                L = L.replace('/layer.tar', '')
                j = read(fmani.replace('manifestJson', L + '/json'))
                md.append(struct_code_block(j, title=str(j['config']['Cmd'])))


def write_markdown():
    app.log.info('------------ WRITE MARKDOWN -------------')
    dl = F.target_dir + '/.da_layers'
    md = ['# Build Documentation']
    md.append('')
    md.append(struct_code_block(dict(os.environ), 'Environment'))
    md.append('')

    for d in os.listdir(dl):
        md.append('## %s' % ' '.join(d.split('_')))
        md.append('')
        document(dl + '/' + d, md)
    s = read(F.write_markdown)
    sep = '<!-- DevApps Autodoc -->'
    if sep in s:
        pre, _, post = s.split(sep, 3)
    else:
        pre, post = '', ''
    md = sep.join((pre, '\n'.join(md), post))
    m = {
        'DA_T_Merged': time.ctime(),
        'DA_ROOT_PW': F.rootpw,
        'DA_BUILD_HOST': socket.gethostname(),
    }
    md = add_metas(md, m)
    with open(F.write_markdown, 'w') as fd:
        fd.write(md)


def add_meta_data():
    app.log.info('------------ ADD METADATA ---------------')
    # breakpoint()
    dl = F.target_dir + '/.da_layers'
    run_cmd('mkdir -p "%s"' % dl)
    i = 1
    for ds in F.dirs:
        dt = dl + ('/%s__' % i) + ds.rsplit('/', 1)[-1]
        run_cmd('mkdir -p "%s"' % dt)
        if exists(ds + '/.da_layers'):
            d1, d2 = (ds + '/.da_layers', dt)
        else:
            d1, d2 = (ds, dt)
        # tar --force-local -cf google.com:files.tar *
        dta = abspath(ds + '.%s.tar' % os.getpid())
        run_cmd('cd "%s" && tar --force-local -cf "%s" .' % (d1, dta))
        run_cmd('cd "%s" && tar --force-local -xf "%s" ' % (d2, dta))
        os.unlink(dta)
        i += 1
    size = os.popen('du "%s"  -d 0' % dl).read()
    app.log.info('Added metadata', count=i, size=size)


def do_dir(ds):
    s = ds.rsplit('/', 1)[-1]
    app.log.info('------------ DO DIR ---------------------%s' % s)
    dt = F.target_dir
    mani = ds + '/manifestJson'
    if exists(mani):
        with open(mani) as fd:
            mani = json.loads(fd.read())[0]
        layers = mani['Layers']
        run_cmd('mkdir -p "%s"' % F.target_dir)
        for l in layers:
            add_layer(ds, l)
    else:
        app.log.info('Merging as plain', dir=ds)
        add_plain_dir(ds)


def add_plain_dir(ds):
    if F.method == 'tar':
        dirs = (ds, F.target_dir)
        run_cmd('(cd "%s" && tar cf - | cd "%s" && tar xf - )' % dirs)
        return
    raise NotImplemented


main = lambda: run_app(run, flags=Flags)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Waiting for a port

"""


import os, sys
from functools import partial
import importlib
from devapp.app import app, do, run_app, system
from devapp.tools import (
    FLG,
    project,
    exists,
    repl_dollar_var_with_env_val,
    now,
    wait_for_port,
)


class Flags:
    autoshort = ''

    class port:
        d = '$http_port'

    class host:
        d = '127.0.0.1'

    class timeout:
        d = 5


# ------------------------------------------------------------------------- end actions
def run():
    t0 = now()
    port = int(repl_dollar_var_with_env_val(FLG.port))
    host = repl_dollar_var_with_env_val(FLG.host)
    timeout = float(FLG.timeout)
    if not do(wait_for_port, port, host, timeout):
        app.die('Timeout')
    app.info('Port available', port=FLG.port, millis=now() - t0)


main = partial(run_app, run, flags=Flags)



import os
import sys

from mdvl import to_md_list


def explain_parse_options():
    """max_explain docstrings can do that passing the file as sys.argv[1]
    In the file we parse the matching occurrance of a switch case statement like
    the one in the max_function file.
    Debug this by printing out - will result in the docstring shown.
    """
    match = sys.argv[1]

    fn = sys.argv[2]
    if not os.path.exists(fn):
        print('Options parser: not found file given by docstring (%s)' % fn)
        return
    with open(fn) as fd:
        s = fd.read()
    block = []
    have = False
    is_match = False
    for line in s.splitlines():
        line = line.strip()
        if match in line:
            is_match = True
        if line == 'esac':
            if is_match:
                have = True
                break
        else:
            if is_match:
                block.append(line)
    if not have or not block:
        print('No case declaration found in', fn)
    # options now like:
    # -d|deactivate)
    # deactivate
    # return $?
    # ;;
    # -w|--welcome)
    keys = []
    shorts = []
    funcs = []
    while len(block) > 1:
        line = block.pop(0)
        if line == '*)':
            continue

        if line.endswith(')'):
            if not '|' in line:
                line = '|' + line
            l = line.split('|', 1)
            shorts.append(l[0].strip())
            keys.append(l[1].split(')', 1)[0].strip())
            while block and 'shift' in block[0]:
                block.pop(0)
            if block:
                funcs.append(block.pop(0))

    out = to_md_list(shorts, keys, funcs)
    print(out)



import json
import os

from devapp.tools import dirname, exists, read_file, write_file

to_run_dir = lambda d, repo, phys_dir: d.replace(phys_dir, repo['run_dir'])


def find_bin_paths(comp_checkout_dir_check_dir):
    comp, cod, d = rcd = comp_checkout_dir_check_dir
    if exists(d + '/bin'):
        comp.setdefault('env_PATH', []).append(to_run_dir(d + '/bin', comp, cod))
    return rcd


def find_py_paths(comp_checkout_dir_check_dir):
    """
    Set PYTHONPATH according to package structure.

    Ok this is tricky. We want, when the comp is later mounted from
    checkout dir to run_dir, have the import statements working, allthough
    not pip installed (we might do this when we have an overlay later, but for
    now we try to be w/o unsharing whenever possible - and it IS possible,
    like in java or ruby, to find our modules just via an env enrichment -
    and that is PYTHONPATH in python.
    Problem though: python's packages declaration is not a declaration but code...
    But Since we have python we can execute that code.
    """
    # TODO: Comment that better
    # cod: checkout dir
    # d: a possible daemon dir:
    comp, cod, d = rcd = comp_checkout_dir_check_dir

    def add(dir_, comp=comp, cod=cod):
        pp = comp.setdefault('env_PYTHONPATH', [])
        pp.append(to_run_dir(dir_, comp, cod))

    if not exists(d):
        return rcd
    # first we check for daemon specific stuff:
    if any([f for f in os.listdir(d) if f.endswith('.py')]):
        add(d)
    fns = d + '/setup.py'
    if not exists(fns):
        return rcd
    # we cache the result of that setup.py import, this might take
    # long, sucks for repeated builds:
    st = os.stat(fns)
    # lets use pycache folder for that, this is gitignored for all sane python projects:
    fnc = d + '/.__pycache__/%s.%s.devapps.json' % (st.st_ino, st.st_mtime)
    if exists(fnc):
        j = read_file(fnc, '')
    else:
        # if this build toolchain would be written in go, they would hate python
        # for this. We have python, so ...piece of cak(e or a) to get the comp stuff:
        j = os.popen(
            '''cd "%s" && python -c 'if 1:
        import setuptools
        import json
        def fake_setup(**kw):
            l = [kw.get(l, "") for l in ("packages", "entry_points", "package_dir")]
            print(json.dumps(l, default=str))
        setuptools.setup = fake_setup
        import setup'
        '''
            % d
        ).read()
        os.makedirs(dirname(fnc)) if not exists(dirname(fnc)) else 0
        write_file(fnc, j)
    if not j:
        return rcd
    pkgs, eps, pkg_dirs = json.loads(j)
    # todo: create the entrypoints
    # for now only the stuff needed to run and that are the pypaths:
    pkgs = [p for p in pkgs if not '.' in p]
    # todo: currently we look only in pkgdirs:
    ds = (pkg_dirs or {'': ''}).values()
    [add(d + '/%s' % v) for v in ds]

    return rcd



# Make coding more python3-ish


__metaclass__ = type


class AXC2Error(Exception):
    pass


class AXC2VaultError(Exception):
    pass


# ------------------------------------------------------ ansible/parsing/vault:
# FIXME: ansible is GPL3, not sure if this is public domain knowhow or not
# Maybe factorize these 100 lines?
import os
import warnings
from binascii import hexlify, unhexlify

HAS_CRYPTOGRAPHY = False
HAS_PYCRYPTO = False
HAS_SOME_PYCRYPTO = False
CRYPTOGRAPHY_BACKEND = None
try:
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', DeprecationWarning)
        from cryptography.exceptions import InvalidSignature
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes, padding
    from cryptography.hazmat.primitives.hmac import HMAC
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
    from cryptography.hazmat.primitives.ciphers import (
        Cipher as C_Cipher,
        algorithms,
        modes,
    )

    CRYPTOGRAPHY_BACKEND = default_backend()
    HAS_CRYPTOGRAPHY = True
except ImportError:
    pass

try:
    from Crypto.Cipher import AES as AES_pycrypto

    HAS_SOME_PYCRYPTO = True

    # Note: Only used for loading obsolete VaultAES files.  All files are written
    # using the newer AES256Vault which does not require md5
    from Crypto.Hash import SHA256 as SHA256_pycrypto
    from Crypto.Hash import HMAC as HMAC_pycrypto

    # Counter import fails for 2.0.1, requires >= 2.6.1 from pip
    from Crypto.Util import Counter as Counter_pycrypto

    # KDF import fails for 2.0.1, requires >= 2.6.1 from pip
    from Crypto.Protocol.KDF import PBKDF2 as PBKDF2_pycrypto

    HAS_PYCRYPTO = True
except ImportError:
    pass


class AES256Vault:

    """
    Vault implementation using AES-CTR with an HMAC-SHA256 authentication code.
    Keys are derived using PBKDF2
    """

    # http://www.daemonology.net/blog/2009-06-11-cryptographic-right-answers.html

    # Note: strings in this class should be byte strings by default.

    def __init__(self):
        if not HAS_CRYPTOGRAPHY and not HAS_PYCRYPTO:
            raise AXC2Error(NEED_CRYPTO_LIBRARY)

    @staticmethod
    def _create_key_cryptography(b_password, b_salt, key_length, iv_length):
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=2 * key_length + iv_length,
            salt=b_salt,
            iterations=10000,
            backend=CRYPTOGRAPHY_BACKEND,
        )
        b_derivedkey = kdf.derive(b_password)

        return b_derivedkey

    @staticmethod
    def _pbkdf2_prf(p, s):
        hash_function = SHA256_pycrypto
        return HMAC_pycrypto.new(p, s, hash_function).digest()

    @classmethod
    def _create_key_pycrypto(cls, b_password, b_salt, key_length, iv_length):

        # make two keys and one iv

        b_derivedkey = PBKDF2_pycrypto(
            b_password,
            b_salt,
            dkLen=(2 * key_length) + iv_length,
            count=10000,
            prf=cls._pbkdf2_prf,
        )
        return b_derivedkey

    @classmethod
    def _gen_key_initctr(cls, b_password, b_salt):
        # 16 for AES 128, 32 for AES256
        key_length = 32

        if HAS_CRYPTOGRAPHY:
            # AES is a 128-bit block cipher, so IVs and counter nonces are 16 bytes
            iv_length = algorithms.AES.block_size // 8

            b_derivedkey = cls._create_key_cryptography(
                b_password, b_salt, key_length, iv_length
            )
            b_iv = b_derivedkey[(key_length * 2) : (key_length * 2) + iv_length]
        elif HAS_PYCRYPTO:
            # match the size used for counter.new to avoid extra work
            iv_length = 16

            b_derivedkey = cls._create_key_pycrypto(
                b_password, b_salt, key_length, iv_length
            )
            b_iv = hexlify(
                b_derivedkey[(key_length * 2) : (key_length * 2) + iv_length]
            )
        else:
            raise AXC2Error(NEED_CRYPTO_LIBRARY + '(Detected in initctr)')

        b_key1 = b_derivedkey[:key_length]
        b_key2 = b_derivedkey[key_length : (key_length * 2)]

        return b_key1, b_key2, b_iv

    @staticmethod
    def _encrypt_cryptography(b_plaintext, b_key1, b_key2, b_iv):
        cipher = C_Cipher(algorithms.AES(b_key1), modes.CTR(b_iv), CRYPTOGRAPHY_BACKEND)
        encryptor = cipher.encryptor()
        padder = padding.PKCS7(algorithms.AES.block_size).padder()
        b_ciphertext = encryptor.update(padder.update(b_plaintext) + padder.finalize())
        b_ciphertext += encryptor.finalize()

        # COMBINE SALT, DIGEST AND DATA
        hmac = HMAC(b_key2, hashes.SHA256(), CRYPTOGRAPHY_BACKEND)
        hmac.update(b_ciphertext)
        b_hmac = hmac.finalize()

        return (
            to_bytes(hexlify(b_hmac), errors='surrogate_or_strict'),
            hexlify(b_ciphertext),
        )

    @staticmethod
    def _encrypt_pycrypto(b_plaintext, b_key1, b_key2, b_iv):
        # PKCS#7 PAD DATA http://tools.ietf.org/html/rfc5652#section-6.3
        bs = AES_pycrypto.block_size
        padding_length = (bs - len(b_plaintext) % bs) or bs
        b_plaintext += to_bytes(
            padding_length * chr(padding_length), encoding='ascii', errors='strict'
        )

        # COUNTER.new PARAMETERS
        # 1) nbits (integer) - Length of the counter, in bits.
        # 2) initial_value (integer) - initial value of the counter. "iv" from _gen_key_initctr

        ctr = Counter_pycrypto.new(128, initial_value=int(b_iv, 16))

        # AES.new PARAMETERS
        # 1) AES key, must be either 16, 24, or 32 bytes long -- "key" from _gen_key_initctr
        # 2) MODE_CTR, is the recommended mode
        # 3) counter=<CounterObject>

        cipher = AES_pycrypto.new(b_key1, AES_pycrypto.MODE_CTR, counter=ctr)

        # ENCRYPT PADDED DATA
        b_ciphertext = cipher.encrypt(b_plaintext)

        # COMBINE SALT, DIGEST AND DATA
        hmac = HMAC_pycrypto.new(b_key2, b_ciphertext, SHA256_pycrypto)

        return (
            to_bytes(hmac.hexdigest(), errors='surrogate_or_strict'),
            hexlify(b_ciphertext),
        )

    @classmethod
    def encrypt(cls, b_plaintext, secret):
        if secret is None:
            raise AXC2VaultError('The secret passed to encrypt() was None')
        b_salt = os.urandom(32)
        # b_password = secret.bytes
        b_password = secret
        b_key1, b_key2, b_iv = cls._gen_key_initctr(b_password, b_salt)

        if HAS_CRYPTOGRAPHY:
            b_hmac, b_ciphertext = cls._encrypt_cryptography(
                b_plaintext, b_key1, b_key2, b_iv
            )
        elif HAS_PYCRYPTO:
            b_hmac, b_ciphertext = cls._encrypt_pycrypto(
                b_plaintext, b_key1, b_key2, b_iv
            )
        else:
            raise AXC2Error(NEED_CRYPTO_LIBRARY + '(Detected in encrypt)')

        b_vaulttext = b'\n'.join([hexlify(b_salt), b_hmac, b_ciphertext])
        # Unnecessary but getting rid of it is a backwards incompatible vault
        # format change
        b_vaulttext = hexlify(b_vaulttext)
        return b_vaulttext

    @classmethod
    def _decrypt_cryptography(cls, b_ciphertext, b_crypted_hmac, b_key1, b_key2, b_iv):
        # b_key1, b_key2, b_iv = self._gen_key_initctr(b_password, b_salt)
        # EXIT EARLY IF DIGEST DOESN'T MATCH
        hmac = HMAC(b_key2, hashes.SHA256(), CRYPTOGRAPHY_BACKEND)
        hmac.update(b_ciphertext)
        try:
            hmac.verify(unhexlify(b_crypted_hmac))
        except InvalidSignature as e:
            raise AXC2VaultError('HMAC verification failed: %s' % e)

        cipher = C_Cipher(algorithms.AES(b_key1), modes.CTR(b_iv), CRYPTOGRAPHY_BACKEND)
        decryptor = cipher.decryptor()
        unpadder = padding.PKCS7(128).unpadder()
        b_plaintext = (
            unpadder.update(decryptor.update(b_ciphertext) + decryptor.finalize())
            + unpadder.finalize()
        )

        return b_plaintext

    @staticmethod
    def _is_equal(b_a, b_b):
        """
        Comparing 2 byte arrrays in constant time
        to avoid timing attacks.

        It would be nice if there was a library for this but
        hey.
        """
        if not (isinstance(b_a, binary_type) and isinstance(b_b, binary_type)):
            raise TypeError('_is_equal can only be used to compare two byte strings')

        # http://codahale.com/a-lesson-in-timing-attacks/
        if len(b_a) != len(b_b):
            return False

        result = 0
        for b_x, b_y in zip(b_a, b_b):
            if PY3:
                result |= b_x ^ b_y
            else:
                result |= ord(b_x) ^ ord(b_y)
        return result == 0

    @classmethod
    def _decrypt_pycrypto(cls, b_ciphertext, b_crypted_hmac, b_key1, b_key2, b_iv):
        # EXIT EARLY IF DIGEST DOESN'T MATCH
        hmac_decrypt = HMAC_pycrypto.new(b_key2, b_ciphertext, SHA256_pycrypto)
        if not cls._is_equal(b_crypted_hmac, to_bytes(hmac_decrypt.hexdigest())):
            return None

        # SET THE COUNTER AND THE CIPHER
        ctr = Counter_pycrypto.new(128, initial_value=int(b_iv, 16))
        cipher = AES_pycrypto.new(b_key1, AES_pycrypto.MODE_CTR, counter=ctr)

        # DECRYPT PADDED DATA
        b_plaintext = cipher.decrypt(b_ciphertext)

        # UNPAD DATA
        if PY3:
            padding_length = b_plaintext[-1]
        else:
            padding_length = ord(b_plaintext[-1])

        b_plaintext = b_plaintext[:-padding_length]
        return b_plaintext

    @classmethod
    def decrypt(cls, b_vaulttext, secret):
        # SPLIT SALT, DIGEST, AND DATA
        b_vaulttext = unhexlify(b_vaulttext)
        b_salt, b_crypted_hmac, b_ciphertext = b_vaulttext.split(b'\n', 2)
        b_salt = unhexlify(b_salt)
        b_ciphertext = unhexlify(b_ciphertext)

        # TODO: would be nice if a VaultSecret could be passed directly to _decrypt_*
        #       (move _gen_key_initctr() to a AES256 VaultSecret or VaultContext impl?)
        # though, likely needs to be python cryptography specific impl that basically
        # creates a Cipher() with b_key1, a Mode.CTR() with b_iv, and a HMAC() with sign key b_key2
        # b_password = secret.bytes
        b_password = secret

        b_key1, b_key2, b_iv = cls._gen_key_initctr(b_password, b_salt)

        if HAS_CRYPTOGRAPHY:
            b_plaintext = cls._decrypt_cryptography(
                b_ciphertext, b_crypted_hmac, b_key1, b_key2, b_iv
            )
        elif HAS_PYCRYPTO:
            b_plaintext = cls._decrypt_pycrypto(
                b_ciphertext, b_crypted_hmac, b_key1, b_key2, b_iv
            )
        else:
            raise AXC2Error(NEED_CRYPTO_LIBRARY + '(Detected in decrypt)')

        return b_plaintext


u8 = lambda s: bytes(s.encode('utf-8'))
u8s = lambda b: b.decode('utf-8')
AES256Vault.encrypt_text = lambda s, pw: u8s(AES256Vault.encrypt(u8(s), u8(pw)))
AES256Vault.decrypt_to_text = lambda cr, pw: u8s(AES256Vault.decrypt(u8(cr), u8(pw)))


class VaultsByCypher:
    AES256 = AES256Vault


def to_bytes(s, *a, **kw):
    return bytes(s)


encrypt = lambda txt, pw: AES256Vault.encrypt_text(txt, pw)
decrypt = lambda txt, pw: AES256Vault.decrypt_to_text(txt, pw)

if __name__ == '__main__':
    v = AES256Vault
    res = v.encrypt(bytes('foo'.encode('utf-8')), b'bar')
    res = v.decrypt(res, b'bar')
    assert res == b'foo'

    res = v.encrypt_text('foo', 'bar')
    assert v.decrypt_to_text(res, 'bar') == 'foo'
